from datetime import datetime, timedelta
from typing import (
    Any,
    Callable,
    Dict,
    Hashable,
    Iterable,
    List,
    Literal,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    Type,
    Union,
)
import typing as t

from pandas._libs import lib
from pandas._libs.tslibs import BaseOffset
from pandas.core.window.indexers import BaseIndexer
import numpy as np
import pandas as pd
import pandas._typing as pdt

from sarus_data_spec.dataspec_validator.parameter_kind import (
    DATASPEC,
    STATIC,
    TRANSFORM,
)
from sarus_data_spec.dataspec_validator.signature import (
    SarusBoundSignature,
    SarusParameter,
    SarusSignature,
    SarusSignatureValue,
)
from sarus_data_spec.dataspec_validator.typing import PEPKind
import sarus_data_spec.typing as st

from ..external_op import ExternalOpImplementation

# Defined in pandas version > 1.3.5
IgnoreRaise = t.Literal["ignore", "raise"]
ValueKeyFunc = Optional[
    Callable[[pd.Series], Union[pd.Series, pdt.AnyArrayLike]]
]
DropKeep = Literal["first", "last", False]
QuantileInterpolation = Literal[
    "linear", "lower", "higher", "midpoint", "nearest"
]
CorrelationMethod = Union[
    Literal["pearson", "kendall", "spearman"],
    Callable[[np.ndarray, np.ndarray], float],
]
MergeHow = Literal["left", "right", "inner", "outer", "cross"]
ArrayConvertible = Union[List, Tuple, pdt.AnyArrayLike]
DatetimeScalar = Union[pdt.Scalar, datetime]
DatetimeScalarOrArrayConvertible = Union[DatetimeScalar, ArrayConvertible]


# ------ CONSTRUCTORS -------
class pd_dataframe(ExternalOpImplementation):
    _transform_id = "pandas.PD_DATAFRAME"
    _signature = SarusSignature(
        SarusParameter(
            name="data",
            annotation=Optional[
                Union[
                    Sequence[Sequence[Any]],
                    Mapping[Hashable, Sequence[Any]],
                    pd.DataFrame,
                ]
            ],
            default=None,
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="index",
            annotation=Optional[pdt.Axes],
            default=None,
        ),
        SarusParameter(
            name="columns",
            annotation=Optional[pdt.Axes],
            default=None,
        ),
        SarusParameter(
            name="dtype",
            annotation=Optional[pdt.Dtype],
            default=None,
        ),
        SarusParameter(
            name="copy",
            annotation=Optional[bool],
            default=None,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        kwargs = signature.collect_kwargs()
        return pd.DataFrame(**kwargs)


class pd_series(ExternalOpImplementation):
    _transform_id = "pandas.PD_SERIES"
    _signature = SarusSignature(
        SarusParameter(
            name="data",
            annotation=Optional[
                Union[pdt.ArrayLike, Iterable, Dict, pdt.Scalar]
            ],
            default=None,
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="index",
            annotation=pd.Index,
            default=None,
        ),
        SarusParameter(
            name="dtype",
            annotation=Optional[pdt.Dtype],
            default=None,
        ),
        SarusParameter(
            name="name",
            annotation=Optional[str],
            default=None,
        ),
        SarusParameter(
            name="copy",
            annotation=bool,
            default=False,
        ),
        SarusParameter(
            name="fastpath",
            annotation=bool,
            default=False,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        kwargs = signature.collect_kwargs()
        return pd.Series(**kwargs)


# ------ DataFrame & Series METHODS ------
class pd_loc(ExternalOpImplementation):
    _transform_id = "pandas.PD_LOC"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=t.Union[pd.Series, pd.DataFrame],
            condition=DATASPEC,
        ),
        SarusParameter(
            name="key",
            annotation=Tuple[Union[str, slice, List[str]], ...],
        ),
    )

    def call(self, signature: SarusSignatureValue) -> pd.DataFrame:
        (this, key) = signature.collect_args()
        return this.loc[key]

    def pep_kind(self, bound_signature: SarusBoundSignature) -> PEPKind:
        """Token preserving if the key is a tuple or a slice that
        selects all the rows (e.g. loc[:, "col"], loc[:]).
        PEP in the other cases.
        """
        key_arg = bound_signature["key"]
        if STATIC.isin(key_arg.parameter_kind()):
            key_value = key_arg.static_value()
            if isinstance(key_value, tuple) and len(key_value) == 2:
                row_key, _ = key_value
                if row_key == slice(None, None, None):
                    return PEPKind.TOKEN_PRESERVING
            elif isinstance(key_value, slice):
                if key_value == slice(None, None, None):
                    return PEPKind.TOKEN_PRESERVING
            elif isinstance(key_value, (int, str)):
                # a scalar selects a single row
                return PEPKind.ROW

        return PEPKind.PEP


class pd_set_loc(ExternalOpImplementation):
    _transform_id = "pandas.PD_SET_LOC"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=Union[pd.DataFrame, pd.Series],
            condition=DATASPEC,
        ),
        SarusParameter(
            name="key",
            annotation=Tuple[Union[str, slice, List[str]], ...],
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="value",
            annotation=t.Any,
            condition=DATASPEC | STATIC,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        (this, key, value) = signature.collect_args()
        this.loc[key] = value
        return this


class pd_iloc(ExternalOpImplementation):
    _transform_id = "pandas.PD_ILOC"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=t.Union[pd.Series, pd.DataFrame],
            condition=DATASPEC,
        ),
        SarusParameter(
            name="key",
            annotation=Tuple[Union[str, slice, List[str]], ...],
        ),
    )

    def call(self, signature: SarusSignatureValue) -> pd.DataFrame:
        (this, key) = signature.collect_args()
        return this.iloc[key]

    def pep_kind(self, bound_signature: SarusBoundSignature) -> PEPKind:
        """Token preserving the alignment if the key is a slice that
        selects all the rows (e.g. iloc[:]).
        PEP in the other cases.
        """
        key_arg = bound_signature["key"]
        if STATIC.isin(key_arg.parameter_kind()):
            key_value = key_arg.static_value()
            if isinstance(key_value, slice):
                if key_value == slice(None, None, None):
                    return PEPKind.TOKEN_PRESERVING
            elif isinstance(key_value, (int, str)):
                # a scalar selects a single row
                return PEPKind.ROW

        return PEPKind.PEP


class pd_set_iloc(ExternalOpImplementation):
    _transform_id = "pandas.PD_SET_ILOC"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=Union[pd.DataFrame, pd.Series],
            condition=DATASPEC,
        ),
        SarusParameter(
            name="key",
            annotation=Tuple[Union[str, slice, List[str]], ...],
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="value",
            annotation=t.Any,
            condition=DATASPEC | STATIC,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        (this, key, value) = signature.collect_args()
        this.iloc[key] = value
        return this


class pd_head(ExternalOpImplementation):
    _transform_id = "pandas.PD_HEAD"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=t.Union[pd.Series, pd.DataFrame],
            condition=DATASPEC,
        ),
        SarusParameter(
            name="n",
            annotation=int,
            default=5,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        (this, kwargs) = signature.collect_kwargs_method()
        return this.head(**kwargs)

    def pep_kind(self, bound_signature: SarusBoundSignature) -> PEPKind:
        return PEPKind.PEP


class pd_astype(ExternalOpImplementation):
    _transform_id = "pandas.PD_ASTYPE"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=t.Union[pd.Series, pd.DataFrame],
            condition=DATASPEC,
        ),
        SarusParameter(
            name="dtype",
            annotation=pdt.Dtype,
        ),
        SarusParameter(
            name="copy",
            annotation=bool,
            default=True,
        ),
        SarusParameter(
            name="errors",
            annotation=IgnoreRaise,
            default="raise",
        ),
    )

    def call(self, signature: SarusSignatureValue) -> pd.DataFrame:
        (this, kwargs) = signature.collect_kwargs_method()
        return this.astype(**kwargs)

    def pep_kind(self, bound_signature: SarusBoundSignature) -> PEPKind:
        return PEPKind.TOKEN_PRESERVING


class pd_getitem(ExternalOpImplementation):
    _transform_id = "pandas.PD_GETITEM"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=t.Union[pd.Series, pd.DataFrame],
            condition=DATASPEC,
        ),
        SarusParameter(
            name="key",
            annotation=t.Any,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        (this, key) = signature.collect_args()
        return this[key]

    def pep_kind(self, bound_signature: SarusBoundSignature) -> PEPKind:
        """PEP in any case. Token preserving if the key is a string or a list
        of strings."""
        if isinstance(bound_signature["key"].static_value(), st.DataSpec):
            key_type = bound_signature["key"].python_type()
            return (
                PEPKind.TOKEN_PRESERVING
                if key_type in [str(str)]
                else PEPKind.PEP
            )
        else:
            key_value = bound_signature["key"].static_value()
            if isinstance(key_value, list):
                # can select columns or rows depending on the type of the list
                # values
                all_strings = all([isinstance(x, str) for x in key_value])
                return PEPKind.TOKEN_PRESERVING if all_strings else PEPKind.PEP
            return (
                PEPKind.TOKEN_PRESERVING
                if isinstance(key_value, str)
                else PEPKind.PEP
            )


class pd_sum(ExternalOpImplementation):
    _transform_id = "pandas.PD_SUM"
    _dp_equivalent_id = "pandas.PD_SUM_DP"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=t.Union[pd.Series, pd.DataFrame],
            condition=DATASPEC,
        ),
        SarusParameter(
            name="axis",
            annotation=t.Optional[pdt.Axis],
            default=None,
        ),
        SarusParameter(
            name="skipna",
            annotation=bool,
            default=True,
        ),
        SarusParameter(
            name="level",
            annotation=t.Optional[pdt.Level],
            default=None,
        ),
        SarusParameter(
            name="numeric_only",
            annotation=t.Optional[bool],
            default=None,
        ),
        SarusParameter(
            name="min_count",
            annotation=int,
            default=0,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        (this, kwargs) = signature.collect_kwargs_method()
        if signature["this"].python_type() == str(pd.Series):
            del kwargs["numeric_only"]
        return this.sum(**kwargs)

    def pep_kind(self, bound_signature: SarusBoundSignature) -> PEPKind:
        """`axis=1`"""
        if bound_signature["this"].python_type() == str(pd.Series):
            return PEPKind.NOT_PEP

        axis = bound_signature["axis"].static_value()
        return PEPKind.TOKEN_PRESERVING if axis == 1 else PEPKind.NOT_PEP


class pd_mean(ExternalOpImplementation):
    _transform_id = "pandas.PD_MEAN"
    _dp_equivalent_id = "pandas.PD_MEAN_DP"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=t.Union[pd.Series, pd.DataFrame],
            condition=DATASPEC,
        ),
        SarusParameter(
            name="axis",
            annotation=int,
            default=0,
        ),
        SarusParameter(
            name="skipna",
            annotation=bool,
            default=True,
        ),
        SarusParameter(
            name="level",
            annotation=t.Optional[pdt.Level],
            default=None,
        ),
        SarusParameter(
            name="numeric_only",
            annotation=bool,
            default=False,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        (this, kwargs) = signature.collect_kwargs_method()
        if signature["this"].python_type() == str(pd.Series):
            del kwargs["numeric_only"]
        return this.mean(**kwargs)

    def pep_kind(self, bound_signature: SarusBoundSignature) -> PEPKind:
        """`axis=1`"""
        if bound_signature["this"].python_type() == str(pd.Series):
            return PEPKind.NOT_PEP

        axis = bound_signature["axis"].static_value()
        return PEPKind.TOKEN_PRESERVING if axis == 1 else PEPKind.NOT_PEP


class pd_std(ExternalOpImplementation):
    _transform_id = "pandas.PD_STD"
    _dp_equivalent_id = "pandas.PD_STD_DP"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=t.Union[pd.Series, pd.DataFrame],
            condition=DATASPEC,
        ),
        SarusParameter(
            name="axis",
            annotation=int,
            default=0,
        ),
        SarusParameter(
            name="skipna",
            annotation=bool,
            default=True,
        ),
        SarusParameter(
            name="level",
            annotation=t.Optional[pdt.Level],
            default=None,
        ),
        SarusParameter(
            name="ddof",
            annotation=int,
            default=1,
        ),
        SarusParameter(
            name="numeric_only",
            annotation=bool,
            default=False,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        (this, kwargs) = signature.collect_kwargs_method()
        if signature["this"].python_type() == str(pd.Series):
            del kwargs["numeric_only"]
        return this.std(**kwargs)

    def pep_kind(self, bound_signature: SarusBoundSignature) -> PEPKind:
        """`axis=1`"""
        axis = bound_signature["axis"].static_value()
        return PEPKind.TOKEN_PRESERVING if axis == 1 else PEPKind.NOT_PEP


class pd_median(ExternalOpImplementation):
    _transform_id = "pandas.PD_MEDIAN"
    _dp_equivalent_id = "pandas.PD_MEDIAN_DP"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=t.Union[pd.Series, pd.DataFrame],
            condition=DATASPEC,
        ),
        SarusParameter(
            name="axis",
            annotation=int,
            default=0,
        ),
        SarusParameter(
            name="skipna",
            annotation=bool,
            default=True,
        ),
        SarusParameter(
            name="level",
            annotation=t.Optional[int],
            default=None,
        ),
        SarusParameter(
            name="numeric_only",
            annotation=bool,
            default=False,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        (this, kwargs) = signature.collect_kwargs_method()
        if signature["this"].python_type() == str(pd.Series):
            del kwargs["numeric_only"]
        return this.median(**kwargs)

    def pep_kind(self, bound_signature: SarusBoundSignature) -> PEPKind:
        """`axis=1`"""
        if bound_signature["this"].python_type() == str(pd.Series):
            return PEPKind.NOT_PEP

        axis = bound_signature["axis"].static_value()
        return PEPKind.TOKEN_PRESERVING if axis == 1 else PEPKind.NOT_PEP


class pd_abs(ExternalOpImplementation):
    _transform_id = "pandas.PD_ABS"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=t.Union[pd.Series, pd.DataFrame],
            condition=DATASPEC,
        )
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this = signature["this"].value
        return this.abs()

    def pep_kind(self, bound_signature: SarusBoundSignature) -> PEPKind:
        return PEPKind.TOKEN_PRESERVING


class pd_drop(ExternalOpImplementation):
    _transform_id = "pandas.PD_DROP"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=t.Union[pd.Series, pd.DataFrame],
            condition=DATASPEC,
        ),
        SarusParameter(
            name="labels",
            annotation=Optional[
                Union[str, List[Union[str, Tuple[str, ...]]], Type[pdt.Level]]
            ],
            default=None,
        ),
        SarusParameter(
            name="axis",
            annotation=Union[int, str],
            default=0,
        ),
        SarusParameter(
            name="index",
            annotation=Optional[Union[pd.Index, List[Union[str, int]]]],
            default=None,
        ),
        SarusParameter(
            name="columns",
            annotation=Optional[Union[pd.Index, List[Union[str, int]]]],
            default=None,
        ),
        SarusParameter(
            name="level",
            annotation=Optional[Union[int, str, Tuple[Union[int, str], ...]]],
            default=None,
        ),
        SarusParameter(
            name="inplace",
            annotation=bool,
            default=False,
            predicate=lambda x: x is False,
        ),
        SarusParameter(
            name="errors",
            annotation=str,
            default="raise",
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, kwargs = signature.collect_kwargs_method()
        if signature["this"].python_type() == str(pd.Series):
            del kwargs["axis"]
        return this.drop(**kwargs)

    def pep_kind(self, bound_signature: SarusBoundSignature) -> PEPKind:
        axis = bound_signature["axis"].static_value()
        if axis in [0, 'columns']:
            return PEPKind.PEP
        else:
            return PEPKind.TOKEN_PRESERVING


class pd_dropna(ExternalOpImplementation):
    _transform_id = "pandas.PD_DROPNA"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=t.Union[pd.Series, pd.DataFrame],
            condition=DATASPEC,
        ),
        SarusParameter(
            name="axis",
            annotation=Optional[Union[int, str]],
            default=0,
        ),
        SarusParameter(
            name="how",
            annotation=Optional[str],
            default="any",
        ),
        SarusParameter(
            name="thresh",
            annotation=Optional[int],
            default=None,
        ),
        SarusParameter(
            name="subset",
            annotation=Optional[Union[pd.Index, List[Union[str, int]]]],
            default=None,
        ),
        SarusParameter(
            name="inplace",
            annotation=bool,
            default=False,
            predicate=lambda x: x is False,
        ),
        name=_transform_id,
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, kwargs = signature.collect_kwargs_method()
        if signature["this"].python_type() == str(pd.Series):
            del kwargs["subset"]
            del kwargs["thresh"]
        return this.dropna(**kwargs)

    def pep_kind(self, bound_signature: SarusBoundSignature) -> PEPKind:
        axis = bound_signature["axis"].static_value()
        if axis in [0, 'columns']:
            return PEPKind.PEP
        else:
            return PEPKind.TOKEN_PRESERVING


class pd_fillna(ExternalOpImplementation):
    _transform_id = "pandas.PD_FILLNA"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=t.Union[pd.Series, pd.DataFrame],
            condition=DATASPEC,
        ),
        SarusParameter(
            name="value",
            annotation=Optional[
                t.Union[pdt.Scalar, Mapping, Sequence, pd.DataFrame]
            ],
            default=None,
        ),
        SarusParameter(
            name="method",
            annotation=Optional[str],
            default=None,
        ),
        SarusParameter(
            name="axis",
            annotation=Optional[pdt.Axis],
            default=None,
        ),
        SarusParameter(
            name="inplace",
            annotation=bool,
            default=False,
            predicate=lambda x: x is False,
        ),
        SarusParameter(
            name="limit",
            annotation=Optional[int],
            default=None,
        ),
        SarusParameter(
            name="downcast",
            annotation=Optional[str],
            default=None,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, kwargs = signature.collect_kwargs_method()
        return this.fillna(**kwargs)

    def pep_kind(self, bound_signature: SarusBoundSignature) -> PEPKind:
        """`method` is `None`"""
        method = bound_signature["method"].static_value()
        if method is None:
            return PEPKind.TOKEN_PRESERVING
        else:
            return PEPKind.NOT_PEP


class pd_isin(ExternalOpImplementation):
    _transform_id = "pandas.PD_ISIN"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=Union[pd.Series, pd.DataFrame],
            condition=DATASPEC,
        ),
        SarusParameter(
            name="values",
            annotation=Union[pd.Series, List[Any], Tuple[Any], pd.DataFrame],
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, kwargs = signature.collect_kwargs_method()
        return this.isin(**kwargs)

    def pep_kind(self, bound_signature: SarusBoundSignature) -> PEPKind:
        return PEPKind.TOKEN_PRESERVING


class pd_isnull(ExternalOpImplementation):
    _transform_id = "pandas.PD_ISNULL"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=Union[pd.Series, pd.DataFrame],
            condition=DATASPEC,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, _ = signature.collect_kwargs_method()
        return this.isnull()

    def pep_kind(self, bound_signature: SarusBoundSignature) -> PEPKind:
        return PEPKind.TOKEN_PRESERVING


class pd_mask(ExternalOpImplementation):
    _transform_id = "pandas.PD_MASK"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=Union[pd.Series, pd.DataFrame],
            condition=DATASPEC,
        ),
        SarusParameter(
            name="cond",
            annotation=Union[pd.Series, pd.DataFrame, pdt.ArrayLike, Callable],
        ),
        SarusParameter(
            name="other",
            annotation=Optional[
                Union[pd.Series, pd.DataFrame, pdt.Scalar, Callable]
            ],
            default=float("nan"),
        ),
        SarusParameter(
            name="inplace",
            annotation=bool,
            default=False,
            predicate=lambda x: x is False,
        ),
        SarusParameter(
            name="axis",
            annotation=Optional[
                Union[int, Literal['index', 'columns', 'rows']]
            ],
            default=None,
        ),
        SarusParameter(
            name="level",
            annotation=Optional[Union[int, str]],
            default=None,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, kwargs = signature.collect_kwargs_method()
        return this.mask(**kwargs)

    def pep_kind(self, bound_signature: SarusBoundSignature) -> PEPKind:
        """`cond` and `other` are not callable"""
        cond = bound_signature["cond"].static_value()
        other = bound_signature["other"].static_value()
        if callable(cond) or callable(other):
            return PEPKind.NOT_PEP
        else:
            return PEPKind.TOKEN_PRESERVING


class pd_notnull(ExternalOpImplementation):
    _transform_id = "pandas.PD_NOTNULL"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=Union[pd.Series, pd.DataFrame],
            condition=DATASPEC,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, _ = signature.collect_kwargs_method()
        return this.notnull()

    def pep_kind(self, bound_signature: SarusBoundSignature) -> PEPKind:
        return PEPKind.TOKEN_PRESERVING


class pd_rename(ExternalOpImplementation):
    _transform_id = "pandas.PD_RENAME"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=Union[pd.DataFrame, pd.Series],
            condition=DATASPEC,
        ),
        SarusParameter(
            name="mapper",
            annotation=Optional[Union[Dict[str, str], Callable[[str], str]]],
            default=None,
        ),
        SarusParameter(
            name="index",
            annotation=Optional[Union[Dict[str, str], Callable[[str], str]]],
            default=None,
        ),
        SarusParameter(
            name="columns",
            annotation=Optional[Union[Dict[str, str], Callable[[str], str]]],
            default=None,
        ),
        SarusParameter(
            name="axis",
            annotation=Optional[Union[int, str]],
            default=None,
        ),
        SarusParameter(
            name='copy',
            annotation=Optional[bool],
            default=None,
        ),
        SarusParameter(
            name="inplace",
            annotation=bool,
            default=False,
            predicate=lambda x: x is False,
        ),
        SarusParameter(
            name='level',
            default=None,
            annotation=Hashable,
        ),
        SarusParameter(
            name="errors",
            annotation=Literal['ignore', 'raise'],
            default='ignore',
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, kwargs = signature.collect_kwargs_method()
        return this.rename(**kwargs)

    def pep_kind(self, bound_signature: SarusBoundSignature) -> PEPKind:
        """`mapper`, `index` and `columns` are not callable"""
        mapper = bound_signature["mapper"].static_value()
        index = bound_signature["index"].static_value()
        columns = bound_signature["columns"].static_value()
        if callable(mapper) or callable(index) or callable(columns):
            return PEPKind.NOT_PEP
        else:
            return PEPKind.TOKEN_PRESERVING


class pd_replace(ExternalOpImplementation):
    _transform_id = "pandas.PD_REPLACE"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=Union[pd.Series, pd.DataFrame],
            condition=DATASPEC,
        ),
        SarusParameter(
            name="to_replace",
            annotation=Union[pdt.Scalar, Mapping, Sequence],
            default=None,
        ),
        SarusParameter(
            name="value",
            annotation=Union[pdt.Scalar, Mapping, Sequence],
        ),
        SarusParameter(
            name="inplace",
            annotation=bool,
            default=False,
            predicate=lambda x: x is False,
        ),
        SarusParameter(
            name="limit",
            annotation=Optional[int],
            default=None,
        ),
        SarusParameter(
            name="regex",
            annotation=bool,
            default=False,
        ),
        SarusParameter(
            name="method",
            annotation=Literal['pad', 'ffill', 'bfill'],
            default=lib.no_default,
        ),
        name=_transform_id,
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, kwargs = signature.collect_kwargs_method()
        return this.replace(**kwargs)

    def pep_kind(self, bound_signature: SarusBoundSignature) -> PEPKind:
        """`value` is not `None`"""
        value = bound_signature["value"].static_value()
        if value is None:
            return PEPKind.NOT_PEP
        else:
            return PEPKind.TOKEN_PRESERVING


class pd_round(ExternalOpImplementation):
    _transform_id = "pandas.PD_ROUND"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=Union[pd.Series, pd.DataFrame],
            condition=DATASPEC,
        ),
        SarusParameter(
            name="decimals",
            annotation=Union[int, dict, pd.Series],
            default=0,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, kwargs = signature.collect_kwargs_method()
        return this.round(**kwargs)

    def pep_kind(self, bound_signature: SarusBoundSignature) -> PEPKind:
        return PEPKind.TOKEN_PRESERVING


class pd_select_dtypes(ExternalOpImplementation):
    _transform_id = "pandas.PD_SELECT_DTYPES"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=t.Union[pd.Series, pd.DataFrame],
            condition=DATASPEC,
        ),
        SarusParameter(
            name="include",
            annotation=t.Optional[t.Union[pdt.Scalar, t.List]],
            default=None,
        ),
        SarusParameter(
            name="exclude",
            annotation=t.Optional[t.Union[pdt.Scalar, t.List]],
            default=None,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        (this, kwargs) = signature.collect_kwargs_method()
        return this.select_dtypes(**kwargs)

    def pep_kind(self, bound_signature: SarusBoundSignature) -> PEPKind:
        return PEPKind.PEP


class pd_add(ExternalOpImplementation):
    _transform_id = "pandas.PD_ADD"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=Union[pd.Series, pd.DataFrame],
            condition=DATASPEC,
        ),
        SarusParameter(
            name="other",
            annotation=Union[pd.Series, pd.DataFrame, pd.Index, float, int],
        ),
        SarusParameter(
            name="fill_value",
            annotation=Optional[Union[float, int]],
            default=None,
        ),
        SarusParameter(
            name="axis",
            annotation=Optional[
                Union[int, Literal['index', 'columns', 'rows']]
            ],
            default='columns',
        ),
        SarusParameter(
            name="level",
            annotation=Optional[Union[int, str]],
            default=None,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, kwargs = signature.collect_kwargs_method()
        if signature["this"].python_type() == str(pd.Series):
            del kwargs["axis"]
        return this.add(**kwargs)

    def pep_kind(self, bound_signature: SarusBoundSignature) -> PEPKind:
        return PEPKind.TOKEN_PRESERVING


# for test only to remove
class pd_add_test(ExternalOpImplementation):
    _transform_id = "pandas.PD_ADD_TEST"
    _pep_equivalent_id = "pandas.PD_ADD_TEST_PEP"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=Union[pd.Series, pd.DataFrame],
            condition=DATASPEC,
        ),
        SarusParameter(
            name="other",
            annotation=Union[pd.Series, pd.DataFrame, pd.Index, float, int],
        ),
        SarusParameter(
            name="fill_value",
            annotation=Optional[Union[float, int]],
            default=None,
        ),
        SarusParameter(
            name="axis",
            annotation=Optional[
                Union[int, Literal['index', 'columns', 'rows']]
            ],
            default='columns',
        ),
        SarusParameter(
            name="level",
            annotation=Optional[Union[int, str]],
            default=None,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, kwargs = signature.collect_kwargs_method()
        if signature["this"].python_type() == str(pd.Series):
            del kwargs["axis"]
        return this.add(**kwargs)

    def pep_kind(self, bound_signature: SarusBoundSignature) -> PEPKind:
        return PEPKind.NOT_PEP


# for test only to remove
class pd_add_test_pep(ExternalOpImplementation):
    _transform_id = "pandas.PD_ADD_TEST_PEP"
    _non_pep_equivalent_id = "pandas.PD_ADD_TEST"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=Union[pd.Series, pd.DataFrame],
            condition=DATASPEC,
        ),
        SarusParameter(
            name="other",
            annotation=Union[pd.Series, pd.DataFrame, pd.Index, float, int],
        ),
        SarusParameter(
            name="fill_value",
            annotation=Optional[Union[float, int]],
            default=None,
        ),
        SarusParameter(
            name="axis",
            annotation=Optional[
                Union[int, Literal['index', 'columns', 'rows']]
            ],
            default='columns',
        ),
        SarusParameter(
            name="level",
            annotation=Optional[Union[int, str]],
            default=None,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, kwargs = signature.collect_kwargs_method()
        if signature["this"].python_type() == str(pd.Series):
            del kwargs["axis"]
        return this.add(**kwargs)

    def pep_kind(self, bound_signature: SarusBoundSignature) -> PEPKind:
        return PEPKind.TOKEN_PRESERVING


class pd_sub(ExternalOpImplementation):
    _transform_id = "pandas.PD_SUB"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=Union[pd.Series, pd.DataFrame],
            condition=DATASPEC,
        ),
        SarusParameter(
            name="other",
            annotation=Union[pd.Series, pd.DataFrame, pd.Index, float, int],
        ),
        SarusParameter(
            name="fill_value",
            annotation=Optional[Union[float, int]],
            default=None,
        ),
        SarusParameter(
            name="axis",
            annotation=Optional[
                Union[int, Literal['index', 'columns', 'rows']]
            ],
            default='columns',
        ),
        SarusParameter(
            name="level",
            annotation=Optional[Union[int, str]],
            default=None,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, kwargs = signature.collect_kwargs_method()
        if signature["this"].python_type() == str(pd.Series):
            del kwargs["axis"]
        return this.sub(**kwargs)

    def pep_kind(self, bound_signature: SarusBoundSignature) -> PEPKind:
        return PEPKind.TOKEN_PRESERVING


class pd_reset_index(ExternalOpImplementation):
    _transform_id = "pandas.PD_RESET_INDEX"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=Union[pd.Series, pd.DataFrame],
            condition=DATASPEC,
        ),
        SarusParameter(
            name="level",
            annotation=pdt.IndexLabel,
            default=None,
        ),
        SarusParameter(
            name="drop",
            annotation=bool,
            default=False,
        ),
        SarusParameter(
            name="inplace",
            annotation=bool,
            default=False,
            predicate=lambda x: x is False,
        ),
        SarusParameter(
            name="col_level",
            annotation=Hashable,
            default=0,
        ),
        SarusParameter(
            name="col_fill",
            annotation=Hashable,
            default="",
        ),
        # > 1.3.5
        # SarusParameter(
        #     name="allow_duplicates",
        #     annotation=Union[bool, lib.NoDefault],
        #     default=lib.no_default,
        # ),
        # SarusParameter(
        #     name="names",
        #     annotation=Optional[Union[Hashable, Sequence[Hashable]]],
        #     default=None,
        # ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, kwargs = signature.collect_kwargs_method()
        if signature["this"].python_type() == str(pd.Series):
            del kwargs["col_level"]
            del kwargs["col_fill"]
        return this.reset_index(**kwargs)


class pd_min(ExternalOpImplementation):
    _transform_id = "pandas.PD_MIN"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=Union[pd.Series, pd.DataFrame],
            condition=DATASPEC,
        ),
        SarusParameter(
            name="axis",
            annotation=Optional[int],
            default=0,
        ),
        SarusParameter(
            name="skipna",
            annotation=bool,
            default=True,
        ),
        SarusParameter(
            name="numeric_only",
            annotation=bool,
            default=False,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, kwargs = signature.collect_kwargs_method()
        return this.min(**kwargs)


class pd_max(ExternalOpImplementation):
    _transform_id = "pandas.PD_MAX"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=Union[pd.Series, pd.DataFrame],
            condition=DATASPEC,
        ),
        SarusParameter(
            name="axis",
            annotation=Optional[int],
            default=0,
        ),
        SarusParameter(
            name="skipna",
            annotation=bool,
            default=True,
        ),
        SarusParameter(
            name="numeric_only",
            annotation=bool,
            default=False,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, kwargs = signature.collect_kwargs_method()
        return this.max(**kwargs)


class pd_shift(ExternalOpImplementation):
    _transform_id = "pandas.PD_SHIFT"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=Union[pd.Series, pd.DataFrame],
            condition=DATASPEC,
        ),
        SarusParameter(
            name="periods",
            annotation=int,
            default=1,
        ),
        SarusParameter(
            name="freq",
            annotation=Optional[pdt.Frequency],
            default=None,
        ),
        SarusParameter(
            name="axis",
            annotation=pdt.Axis,
            default=0,
        ),
        SarusParameter(
            name="fill_value",
            annotation=Hashable,
            default=lib.no_default,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, kwargs = signature.collect_kwargs_method()
        return this.shift(**kwargs)


class pd_any(ExternalOpImplementation):
    _transform_id = "pandas.PD_ANY"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=Union[pd.Series, pd.DataFrame],
            condition=DATASPEC,
        ),
        SarusParameter(
            name="axis",
            annotation=pdt.Axis,
            default=0,
        ),
        SarusParameter(
            name="bool_only",
            annotation=Optional[bool],
            default=None,
        ),
        SarusParameter(
            name="skipna",
            annotation=bool,
            default=True,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, kwargs = signature.collect_kwargs_method()
        return this.any(**kwargs)


class pd_describe(ExternalOpImplementation):
    _transform_id = "pandas.PD_DESCRIBE"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=Union[pd.Series, pd.DataFrame],
            condition=DATASPEC,
        ),
        SarusParameter(
            name="percentiles",
            annotation=Optional[Sequence[float]],
            default=None,
        ),
        SarusParameter(
            name="include",
            annotation=Optional[Union[Literal['all'], List[pdt.Dtype]]],
            default=None,
        ),
        SarusParameter(
            name="exclude",
            annotation=Optional[List[pdt.Dtype]],
            default=None,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, kwargs = signature.collect_kwargs_method()
        return this.describe(**kwargs)


class pd_quantile(ExternalOpImplementation):
    _transform_id = "pandas.PD_QUANTILE"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=Union[pd.Series, pd.DataFrame],
            condition=DATASPEC,
        ),
        SarusParameter(
            name="q",
            annotation=Union[float, pdt.AnyArrayLike, Sequence[float]],
            default=0.5,
        ),
        SarusParameter(
            name="axis",
            annotation=pdt.Axis,
            default=0,
        ),
        SarusParameter(
            name="numeric_only",
            annotation=bool,
            default=False,
        ),
        SarusParameter(
            name="interpolation",
            annotation=QuantileInterpolation,
            default="linear",
        ),
        # > 1.3.5
        # SarusParameter(
        #     name="method",
        #     annotation=Literal["single", "table"],
        #     default="single",
        # ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, kwargs = signature.collect_kwargs_method()
        if signature["this"].python_type() == str(pd.Series):
            del kwargs["axis"]
            del kwargs["numeric_only"]
        return this.quantile(**kwargs)


class pd_reindex(ExternalOpImplementation):
    _transform_id = "pandas.PD_REINDEX"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=Union[pd.Series, pd.DataFrame],
            condition=DATASPEC,
        ),
        SarusParameter(
            name="labels",
            annotation=Optional[pdt.ArrayLike],
            default=None,
        ),
        SarusParameter(
            name="index",
            annotation=Optional[pdt.ArrayLike],
            default=None,
        ),
        SarusParameter(
            name="columns",
            annotation=Optional[pdt.ArrayLike],
            default=None,
        ),
        SarusParameter(
            name="axis",
            annotation=Optional[pdt.Axis],
            default=None,
        ),
        SarusParameter(
            name="method",
            annotation=Optional[
                Literal["backfill", "bfill", "pad", "ffill", "nearest"]
            ],
            default=None,
        ),
        SarusParameter(
            name="copy",
            annotation=bool,
            default=True,
        ),
        SarusParameter(
            name="level",
            annotation=Optional[pdt.Level],
            default=None,
        ),
        SarusParameter(
            name="fill_value",
            annotation=Optional[pdt.Scalar],
            default=np.nan,
        ),
        SarusParameter(
            name="limit",
            annotation=Optional[int],
            default=None,
        ),
        SarusParameter(
            name="tolerance",
            annotation=Optional[Union[pdt.Scalar, List[pdt.Scalar]]],
            default=None,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, kwargs = signature.collect_kwargs_method()
        # The pandas method is a bit weird and allows only one of
        # these to be specified at a time
        for name in ["labels", "index", "columns", "axis"]:
            if kwargs[name] is None:
                del kwargs[name]
        return this.reindex(**kwargs)


class pd_count(ExternalOpImplementation):
    _transform_id = "pandas.PD_COUNT"
    _dp_equivalent_id = "pandas.PD_COUNT_DP"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=Union[pd.Series, pd.DataFrame],
            condition=DATASPEC,
        ),
        SarusParameter(
            name="axis",
            annotation=pdt.Axis,
            default=0,
        ),
        SarusParameter(
            name="numeric_only",
            annotation=bool,
            default=False,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, kwargs = signature.collect_kwargs_method()
        if signature["this"].python_type() == str(pd.Series):
            del kwargs["axis"]
            del kwargs["numeric_only"]
        return this.count(**kwargs)


class pd_transpose(ExternalOpImplementation):
    _transform_id = "pandas.PD_TRANSPOSE"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=Union[pd.Series, pd.DataFrame],
            condition=DATASPEC,
        ),
        SarusParameter(
            name="copy",
            annotation=bool,
            default=False,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, kwargs = signature.collect_kwargs_method()
        return this.transpose(**kwargs)


class pd_value_counts(ExternalOpImplementation):
    _transform_id = "pandas.PD_VALUE_COUNTS"
    _dp_equivalent_id = "pandas.PD_VALUE_COUNTS_DP"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=Union[pd.Series, pd.DataFrame],
            condition=DATASPEC,
        ),
        SarusParameter(
            name="subset",
            annotation=Optional[Sequence[Hashable]],
            default=None,
        ),
        SarusParameter(
            name="normalize",
            annotation=bool,
            default=False,
        ),
        SarusParameter(
            name="sort",
            annotation=bool,
            default=True,
        ),
        SarusParameter(
            name="ascending",
            annotation=bool,
            default=False,
        ),
        SarusParameter(
            name="dropna",
            annotation=bool,
            default=True,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, kwargs = signature.collect_kwargs_method()
        if signature["this"].python_type() == str(pd.Series):
            del kwargs["subset"]
        return this.value_counts(**kwargs)


class pd_to_dict(ExternalOpImplementation):
    _transform_id = "pandas.PD_TO_DICT"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=pd.DataFrame,
            condition=DATASPEC,
        ),
        SarusParameter(
            name="orient",
            annotation=Literal[
                "dict", "list", "series", "split", "tight", "records", "index"
            ],
            default="dict",
        ),
        SarusParameter(
            name="into",
            annotation=Type[dict],
            default=dict,
        ),
        # > 1.3.5
        # SarusParameter(
        #     name="index",
        #     annotation=bool,
        #     default=True,
        # ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, kwargs = signature.collect_kwargs_method()
        return this.to_dict(**kwargs)


class pd_apply(ExternalOpImplementation):
    _transform_id = "pandas.PD_APPLY"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=Union[pd.DataFrame, pd.Series],
            condition=DATASPEC,
        ),
        SarusParameter(
            name="func",
            annotation=pdt.AggFuncTypeBase,
            condition=STATIC | TRANSFORM,
            predicate=lambda x: isinstance(x, str),
        ),
        SarusParameter(
            name="axis",
            annotation=pdt.Axis,
            default=0,
        ),
        SarusParameter(
            name="raw",
            annotation=bool,
            default=False,
        ),
        SarusParameter(
            name="result_type",
            annotation=Optional[Literal["expand", "reduce", "broadcast"]],
            default=None,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, kwargs = signature.collect_kwargs_method()
        if signature["this"].python_type() == str(pd.Series):
            del kwargs["axis"]
            del kwargs["result_type"]
            del kwargs["raw"]
        return this.apply(**kwargs)


class pd_applymap(ExternalOpImplementation):
    _transform_id = "pandas.PD_APPLYMAP"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=pd.DataFrame,
            condition=DATASPEC,
        ),
        SarusParameter(
            name="func",
            annotation=pdt.AggFuncTypeBase,
            condition=STATIC | TRANSFORM,
            predicate=lambda x: isinstance(x, str),
        ),
        SarusParameter(
            name="na_action",
            annotation=Optional[Literal["ignore"]],
            default=None,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, kwargs = signature.collect_kwargs_method()
        return this.applymap(**kwargs)

    def pep_kind(self, bound_signature: SarusBoundSignature) -> PEPKind:
        return PEPKind.TOKEN_PRESERVING


class pd_map(ExternalOpImplementation):
    _transform_id = "pandas.PD_MAP"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=pd.Series,
            condition=DATASPEC,
        ),
        SarusParameter(
            name="arg",
            annotation=pdt.AggFuncTypeBase,
            condition=STATIC | TRANSFORM,
            predicate=lambda x: isinstance(x, str),
        ),
        SarusParameter(
            name="na_action",
            annotation=Optional[Literal["ignore"]],
            default=None,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, kwargs = signature.collect_kwargs_method()
        return this.map(**kwargs)

    def pep_kind(self, bound_signature: SarusBoundSignature) -> PEPKind:
        return PEPKind.TOKEN_PRESERVING


class pd_skew(ExternalOpImplementation):
    _transform_id = "pandas.PD_SKEW"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=Union[pd.DataFrame, pd.Series],
            condition=DATASPEC,
        ),
        SarusParameter(
            name="axis",
            annotation=Optional[pdt.Axis],
            default=0,
        ),
        SarusParameter(
            name="skipna",
            annotation=bool,
            default=True,
        ),
        SarusParameter(
            name="numeric_only",
            annotation=bool,
            default=False,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, kwargs = signature.collect_kwargs_method()
        return this.skew(**kwargs)


class pd_kurtosis(ExternalOpImplementation):
    _transform_id = "pandas.PD_KURTOSIS"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=Union[pd.DataFrame, pd.Series],
            condition=DATASPEC,
        ),
        SarusParameter(
            name="axis",
            annotation=Optional[pdt.Axis],
            default=0,
        ),
        SarusParameter(
            name="skipna",
            annotation=bool,
            default=True,
        ),
        SarusParameter(
            name="numeric_only",
            annotation=bool,
            default=False,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, kwargs = signature.collect_kwargs_method()
        return this.kurtosis(**kwargs)


class pd_agg(ExternalOpImplementation):
    _transform_id = "pandas.PD_AGG"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=Union[pd.DataFrame, pd.Series],
            condition=DATASPEC,
        ),
        SarusParameter(
            name="func",
            annotation=Optional[pdt.AggFuncTypeBase],
            default=None,
        ),
        SarusParameter(
            name="axis",
            annotation=pdt.Axis,
            default=0,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, kwargs = signature.collect_kwargs_method()
        return this.agg(**kwargs)


class pd_droplevel(ExternalOpImplementation):
    _transform_id = "pandas.PD_DROPLEVEL"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=Union[pd.DataFrame, pd.Series],
            condition=DATASPEC,
        ),
        SarusParameter(
            name="level",
            annotation=pd.Index,
        ),
        SarusParameter(
            name="axis",
            annotation=pdt.Axis,
            default=0,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, kwargs = signature.collect_kwargs_method()
        return this.droplevel(**kwargs)


class pd_sort_values(ExternalOpImplementation):
    _transform_id = "pandas.PD_SORT_VALUES"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=pd.DataFrame,
            condition=DATASPEC,
        ),
        SarusParameter(
            name="by",
            annotation=pdt.IndexLabel,
        ),
        SarusParameter(
            name="axis",
            annotation=pdt.Axis,
            default=0,
        ),
        SarusParameter(
            name="ascending",
            annotation=Union[bool, List[bool], Tuple[bool, ...]],
            default=True,
        ),
        SarusParameter(
            name="inplace",
            annotation=bool,
            default=False,
            predicate=lambda x: x is False,
        ),
        SarusParameter(
            name="kind",
            annotation=str,
            default="quicksort",
        ),
        SarusParameter(
            name="na_position",
            annotation=str,
            default="last",
        ),
        SarusParameter(
            name="ignore_index",
            annotation=bool,
            default=False,
        ),
        SarusParameter(
            name="key",
            annotation=ValueKeyFunc,
            default=None,
        ),
        name=_transform_id,
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, kwargs = signature.collect_kwargs_method()
        return this.sort_values(**kwargs)


class pd_sort_values_series(ExternalOpImplementation):
    _transform_id = "pandas.PD_SORT_VALUES_SERIES"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=pd.Series,
            condition=DATASPEC,
        ),
        SarusParameter(
            name="ascending",
            annotation=Union[bool, List[bool], Tuple[bool, ...]],
            default=True,
        ),
        SarusParameter(
            name="inplace",
            annotation=bool,
            default=False,
            predicate=lambda x: x is False,
        ),
        SarusParameter(
            name="kind",
            annotation=str,
            default="quicksort",
        ),
        SarusParameter(
            name="na_position",
            annotation=str,
            default="last",
        ),
        SarusParameter(
            name="ignore_index",
            annotation=bool,
            default=False,
        ),
        SarusParameter(
            name="key",
            annotation=ValueKeyFunc,
            default=None,
        ),
        name=_transform_id,
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, kwargs = signature.collect_kwargs_method()
        return this.sort_values(**kwargs)


class pd_drop_duplicates(ExternalOpImplementation):
    _transform_id = "pandas.PD_DROP_DUPLICATES"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=Union[pd.DataFrame, pd.Series],
            condition=DATASPEC,
        ),
        SarusParameter(
            name="subset",
            annotation=Optional[Union[Hashable, Sequence[Hashable]]],
            default=None,
        ),
        SarusParameter(
            name="keep",
            annotation=DropKeep,
            default="first",
        ),
        SarusParameter(
            name="inplace",
            annotation=bool,
            default=False,
            predicate=lambda x: x is False,
        ),
        SarusParameter(
            name="ignore_index",
            annotation=bool,
            default=False,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, kwargs = signature.collect_kwargs_method()
        if signature["this"].python_type() == str(pd.Series):
            del kwargs["subset"]
            del kwargs["ignore_index"]
        return this.drop_duplicates(**kwargs)


class pd_corr(ExternalOpImplementation):
    _transform_id = "pandas.PD_CORR"
    _dp_equivalent_id = "pandas.PD_CORR_DP"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=pd.DataFrame,
            condition=DATASPEC,
        ),
        SarusParameter(
            name="method",
            annotation=CorrelationMethod,
            default="pearson",
        ),
        SarusParameter(
            name="min_periods",
            annotation=int,
            default=1,
        ),
        # > 1.3.5
        # SarusParameter(
        #     name="numeric_only",
        #     annotation=bool,
        #     default=False,
        # ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, kwargs = signature.collect_kwargs_method()
        return this.corr(**kwargs)


class pd_corr_series(ExternalOpImplementation):
    _transform_id = "pandas.PD_CORR_SERIES"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=pd.Series,
        ),
        SarusParameter(
            name="other",
            annotation=pd.Series,
        ),
        SarusParameter(
            name="method",
            annotation=CorrelationMethod,
            default="pearson",
        ),
        SarusParameter(
            name="min_periods",
            annotation=int,
            default=1,
        ),
        # > 1.3.5
        # SarusParameter(
        #     name="numeric_only",
        #     annotation=bool,
        #     default=False,
        # ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, kwargs = signature.collect_kwargs_method()
        return this.corr(**kwargs)


# ------ DataFrame & Series PROPERTIES ------
class pd_shape(ExternalOpImplementation):
    _transform_id = "pandas.PD_SHAPE"
    _dp_equivalent_id = "pandas.PD_SHAPE_DP"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=t.Union[pd.Series, pd.DataFrame],
            condition=DATASPEC,
        )
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        (this,) = signature.collect_args()
        return this.shape


class pd_ndim(ExternalOpImplementation):
    _transform_id = "pandas.PD_NDIM"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=Union[pd.DataFrame, pd.Series],
            condition=DATASPEC,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        (this,) = signature.collect_args()
        return this.ndim


class pd_name(ExternalOpImplementation):
    _transform_id = "pandas.PD_NAME"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=pd.Series,
            condition=DATASPEC,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        (this,) = signature.collect_args()
        return this.name


class pd_size(ExternalOpImplementation):
    _transform_id = "pandas.PD_SIZE"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=Union[pd.DataFrame, pd.Series],
            condition=DATASPEC,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        (this,) = signature.collect_args()
        return this.size


class pd_axes(ExternalOpImplementation):
    _transform_id = "pandas.PD_AXES"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=pd.DataFrame,
            condition=DATASPEC,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        (this,) = signature.collect_args()
        return this.axes


class pd_columns(ExternalOpImplementation):
    _transform_id = "pandas.PD_COLUMNS"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=pd.DataFrame,
            condition=DATASPEC,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        (this,) = signature.collect_args()
        return this.columns


class pd_index(ExternalOpImplementation):
    _transform_id = "pandas.PD_INDEX"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=Union[pd.DataFrame, pd.Series],
            condition=DATASPEC,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        (this,) = signature.collect_args()
        return this.index


class pd_dtype(ExternalOpImplementation):
    _transform_id = "pandas.PD_DTYPE"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=pd.Series,
            condition=DATASPEC,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        (this,) = signature.collect_args()
        return this.dtype


class pd_dtypes(ExternalOpImplementation):
    _transform_id = "pandas.PD_DTYPES"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=Union[pd.DataFrame, pd.Series],
            condition=DATASPEC,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        (this,) = signature.collect_args()
        return this.dtypes


class pd_values(ExternalOpImplementation):
    _transform_id = "pandas.PD_VALUES"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=Union[pd.DataFrame, pd.Series],
            condition=DATASPEC,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        (this,) = signature.collect_args()
        return this.values


class pd_join(ExternalOpImplementation):
    _transform_id = "pandas.PD_JOIN"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=pd.DataFrame,
            condition=DATASPEC,
        ),
        SarusParameter(
            name="other",
            annotation=Union[pd.DataFrame, pd.Series],
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="on",
            annotation=Optional[pdt.IndexLabel],
            default=None,
        ),
        SarusParameter(
            name="how",
            annotation=MergeHow,
            default="left",
        ),
        SarusParameter(
            name="lsuffix",
            annotation=str,
            default="",
        ),
        SarusParameter(
            name="rsuffix",
            annotation=str,
            default="",
        ),
        SarusParameter(
            name="sort",
            annotation=bool,
            default=False,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, kwargs = signature.collect_kwargs_method()
        return this.join(**kwargs)


class pd_groupby(ExternalOpImplementation):
    _transform_id = "pandas.PD_GROUPBY"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=pd.DataFrame,
            condition=DATASPEC,
        ),
        SarusParameter(
            name="by",
            annotation=Union[Mapping, Callable, str, List[str], Tuple[str]],
            default=None,
        ),
        SarusParameter(
            name="axis",
            annotation=pdt.Axis,
            default=0,
        ),
        SarusParameter(
            name="level",
            annotation=Optional[pdt.IndexLabel],
            default=None,
        ),
        SarusParameter(
            name="as_index",
            annotation=bool,
            default=True,
        ),
        SarusParameter(
            name="sort",
            annotation=bool,
            default=True,
        ),
        SarusParameter(
            name="group_keys",
            annotation=bool,
            default=True,
        ),
        SarusParameter(
            name="observed",
            annotation=bool,
            default=False,
        ),
        SarusParameter(
            name="dropna",
            annotation=bool,
            default=True,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, kwargs = signature.collect_kwargs_method()
        return this.groupby(**kwargs)

    def pep_kind(self, signature: SarusBoundSignature) -> PEPKind:
        by = signature["axis"].static_value()
        if callable(by):
            return PEPKind.NOT_PEP
        else:
            return PEPKind.PEP


class pd_merge(ExternalOpImplementation):
    _transform_id = "pandas.PD_MERGE"
    _signature = SarusSignature(
        SarusParameter(
            name="left",
            annotation=pd.DataFrame,
            condition=DATASPEC,
        ),
        SarusParameter(
            name="right",
            annotation=Union[pd.DataFrame, pd.Series],
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="how",
            annotation=MergeHow,
            default="inner",
        ),
        SarusParameter(
            name="on",
            annotation=Optional[Union[pdt.IndexLabel, List[pdt.IndexLabel]]],
            default=None,
        ),
        SarusParameter(
            name="left_on",
            annotation=Optional[Union[pdt.IndexLabel, List[pdt.IndexLabel]]],
            default=None,
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="right_on",
            annotation=Optional[Union[pdt.IndexLabel, List[pdt.IndexLabel]]],
            default=None,
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="left_index",
            annotation=bool,
            default=False,
        ),
        SarusParameter(
            name="right_index",
            annotation=bool,
            default=False,
        ),
        SarusParameter(
            name="sort",
            annotation=bool,
            default=False,
        ),
        SarusParameter(
            name="suffixes",
            annotation=Tuple[str, str],
            default=("_x", "_y"),
        ),
        SarusParameter(
            name="copy",
            annotation=bool,
            default=True,
        ),
        SarusParameter(
            name="indicator",
            annotation=bool,
            default=False,
        ),
        SarusParameter(
            name="validate",
            annotation=Optional[str],
            default=None,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, kwargs = signature.collect_kwargs_method()
        return this.merge(**kwargs)


class pd_append(ExternalOpImplementation):
    _transform_id = "pandas.PD_APPEND"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=pd.DataFrame,
            condition=DATASPEC,
        ),
        SarusParameter(
            name="other",
            annotation=Union[pd.DataFrame, pd.Series],
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="ignore_index",
            annotation=bool,
            default=False,
        ),
        SarusParameter(
            name="verify_integrity",
            annotation=bool,
            default=False,
        ),
        SarusParameter(
            name="sort",
            annotation=bool,
            default=False,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, kwargs = signature.collect_kwargs_method()
        return this.append(**kwargs)


class pd_nunique(ExternalOpImplementation):
    _transform_id = "pandas.PD_NUNIQUE"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=Union[pd.DataFrame, pd.Series],
            condition=DATASPEC,
        ),
        SarusParameter(
            name="axis",
            annotation=pdt.Axis,
            default=0,
        ),
        SarusParameter(
            name="dropna",
            annotation=bool,
            default=True,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, kwargs = signature.collect_kwargs_method()
        if signature["this"].python_type() == str(pd.Series):
            del kwargs["axis"]
        return this.nunique(**kwargs)


class pd_unique(ExternalOpImplementation):
    _transform_id = "pandas.PD_UNIQUE"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=pd.Series,
            condition=DATASPEC,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        (this,) = signature.collect_args()
        return this.unique()


class pd_rolling(ExternalOpImplementation):
    _transform_id = "pandas.PD_ROLLING"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=Union[pd.DataFrame, pd.Series],
            condition=DATASPEC,
        ),
        SarusParameter(
            name="window",
            annotation=Union[int, timedelta, BaseOffset, BaseIndexer],
        ),
        SarusParameter(
            name="min_periods",
            annotation=Optional[int],
            default=None,
        ),
        SarusParameter(
            name="center",
            annotation=bool,
            default=False,
        ),
        SarusParameter(
            name="win_type",
            annotation=Optional[str],
            default=None,
        ),
        SarusParameter(
            name="on",
            annotation=Optional[str],
            default=None,
        ),
        SarusParameter(
            name="axis",
            annotation=pdt.Axis,
            default=0,
        ),
        SarusParameter(
            name="closed",
            annotation=Optional[str],
            default=None,
        ),
        SarusParameter(
            name="method",
            annotation=str,
            default="single",
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, kwargs = signature.collect_kwargs_method()
        return this.rolling(**kwargs)


# ------ FUNCTIONS ------
class pd_eq(ExternalOpImplementation):
    _transform_id = "pandas.PD_EQ"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=Union[pd.DataFrame, pd.Series],
            condition=DATASPEC,
        ),
        SarusParameter(
            name="other",
            annotation=Union[pd.DataFrame, pd.Series],
            condition=DATASPEC | STATIC,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        (this, other) = signature.collect_args()
        return this == other


class pd_concat(ExternalOpImplementation):
    _transform_id = "pandas.PD_CONCAT"
    _signature = SarusSignature(
        SarusParameter(
            name="objs",
            annotation=Union[
                Iterable[pd.core.generic.NDFrame],
                Mapping[Hashable, pd.core.generic.NDFrame],
            ],
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="axis",
            annotation=pdt.Axis,
            default=0,
        ),
        SarusParameter(
            name="join",
            annotation=str,
            default="outer",
        ),
        SarusParameter(
            name="ignore_index",
            annotation=bool,
            default=False,
        ),
        SarusParameter(
            name="keys",
            annotation=Optional[Any],
            default=None,
        ),
        SarusParameter(
            name="levels",
            annotation=Optional[Any],
            default=None,
        ),
        SarusParameter(
            name="names",
            annotation=Optional[Any],
            default=None,
        ),
        SarusParameter(
            name="verify_integrity",
            annotation=bool,
            default=False,
        ),
        SarusParameter(
            name="sort",
            annotation=bool,
            default=False,
        ),
        SarusParameter(
            name="copy",
            annotation=Optional[bool],
            default=None,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        kwargs = signature.collect_kwargs()
        return pd.concat(**kwargs)


class pd_get_dummies(ExternalOpImplementation):
    _transform_id = "pandas.PD_GET_DUMMIES"
    _signature = SarusSignature(
        SarusParameter(
            name="data",
            annotation=Union[pd.DataFrame, pd.Series],
            condition=DATASPEC,
        ),
        SarusParameter(
            name="prefix",
            annotation=Optional[Union[str, Iterable[str], Dict[str, str]]],
            default=None,
        ),
        SarusParameter(
            name="prefix_sep",
            annotation=Union[str, Iterable[str], Dict[str, str]],
            default="_",
        ),
        SarusParameter(
            name="dummy_na",
            annotation=bool,
            default=False,
        ),
        SarusParameter(
            name="columns",
            annotation=Optional[List[str]],
            default=None,
        ),
        SarusParameter(
            name="sparse",
            annotation=bool,
            default=False,
        ),
        SarusParameter(
            name="drop_first",
            annotation=bool,
            default=False,
        ),
        SarusParameter(
            name="dtype",
            annotation=Optional[np.dtype],
            default=None,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> pd.DataFrame:
        kwargs = signature.collect_kwargs()
        return pd.get_dummies(**kwargs)


class pd_to_datetime(ExternalOpImplementation):
    _transform_id = "pandas.TO_DATETIME"
    _signature = SarusSignature(
        SarusParameter(
            name="arg",
            annotation=DatetimeScalarOrArrayConvertible,
            condition=DATASPEC,
        ),
        SarusParameter(
            name="errors",
            annotation=str,
            default="raise",
        ),
        SarusParameter(
            name="dayfirst",
            annotation=bool,
            default=False,
        ),
        SarusParameter(
            name="yearfirst",
            annotation=bool,
            default=False,
        ),
        SarusParameter(
            name="utc",
            annotation=bool,
            default=False,
        ),
        SarusParameter(
            name="format",
            annotation=Optional[str],
            default=None,
        ),
        SarusParameter(
            name="exact",
            annotation=Union[bool, lib.NoDefault],
            default=lib.no_default,
        ),
        SarusParameter(
            name="unit",
            annotation=Optional[str],
            default=None,
        ),
        SarusParameter(
            name="infer_datetime_format",
            annotation=Union[bool, lib.NoDefault],
            default=lib.no_default,
        ),
        SarusParameter(
            name="origin",
            annotation=str,
            default="unix",
        ),
        SarusParameter(
            name="cache",
            annotation=bool,
            default=True,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        kwargs = signature.collect_kwargs()
        return pd.to_datetime(**kwargs)


# ------ INDEX METHODS ------
class pd_union(ExternalOpImplementation):
    _transform_id = "pandas.PD_UNION"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=pd.Index,
            condition=DATASPEC,
        ),
        SarusParameter(
            name="other",
            annotation=Union[pd.Index, pdt.ArrayLike],
            condition=DATASPEC | STATIC,
        ),
        SarusParameter(
            name="sort",
            annotation=Optional[bool],
            default=None,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, kwargs = signature.collect_kwargs_method()
        return this.union(**kwargs)



from typing import Any, List, Optional, Tuple, Union

from sarus_data_spec.dataspec_validator.signature import (
    SarusParameter,
    SarusSignature,
    SarusSignatureValue,
)

from ..external_op import ExternalOpImplementation

try:
    from sklearn import pipeline
    from sklearn.base import BaseEstimator
except ModuleNotFoundError:
    BaseEstimator = Any


class sk_pipeline(ExternalOpImplementation):
    _transform_id = "sklearn.SK_PIPELINE"
    _signature = SarusSignature(
        SarusParameter(
            name="steps",
            annotation=List[Tuple[str, BaseEstimator]],
        ),
        SarusParameter(
            name="memory",
            annotation=Optional[Union[str, Any]],
            default=None,
            predicate=lambda x: x is None,
        ),
        SarusParameter(
            name="verbose",
            annotation=bool,
            default=False,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        kwargs = signature.collect_kwargs()
        return pipeline.Pipeline(**kwargs)



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""
import builtins
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.message
import sarus_data_spec.protobuf.statistics_pb2
import typing
import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor = ...

class Size(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
    class PropertiesEntry(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        key: typing.Text = ...
        value: typing.Text = ...
        def __init__(self,
            *,
            key : typing.Text = ...,
            value : typing.Text = ...,
            ) -> None: ...
        def ClearField(self, field_name: typing_extensions.Literal["key",b"key","value",b"value"]) -> None: ...

    UUID_FIELD_NUMBER: builtins.int
    DATASET_FIELD_NUMBER: builtins.int
    NAME_FIELD_NUMBER: builtins.int
    STATISTICS_FIELD_NUMBER: builtins.int
    PROPERTIES_FIELD_NUMBER: builtins.int
    uuid: typing.Text = ...
    dataset: typing.Text = ...
    """uuid"""

    name: typing.Text = ...
    @property
    def statistics(self) -> sarus_data_spec.protobuf.statistics_pb2.Statistics: ...
    @property
    def properties(self) -> google.protobuf.internal.containers.ScalarMap[typing.Text, typing.Text]: ...
    def __init__(self,
        *,
        uuid : typing.Text = ...,
        dataset : typing.Text = ...,
        name : typing.Text = ...,
        statistics : typing.Optional[sarus_data_spec.protobuf.statistics_pb2.Statistics] = ...,
        properties : typing.Optional[typing.Mapping[typing.Text, typing.Text]] = ...,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["statistics",b"statistics"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["dataset",b"dataset","name",b"name","properties",b"properties","statistics",b"statistics","uuid",b"uuid"]) -> None: ...
global___Size = Size



from __future__ import annotations

from enum import Enum
from typing import Collection, List, Optional, Protocol
import logging

from sarus_data_spec.storage.typing import Storage
import sarus_data_spec.typing as st

logger = logging.getLogger(__name__)


class DataspecPrivacyPolicy(Enum):
    WHITE_LISTED = "Whitelisted"
    DP = "Differentially-private evaluation"
    SYNTHETIC = "Evaluated from synthetic data only"
    PUBLIC = "Public"


class PEPKind(Enum):
    NOT_PEP = 0
    PEP = 1
    TOKEN_PRESERVING = 2
    ROW = 3


class DataspecValidator(Protocol):
    def storage(self) -> Storage:
        ...

    def verifies(
        self,
        variant_constraint: st.VariantConstraint,
        kind: st.ConstraintKind,
        public_context: Collection[str],
        privacy_limit: Optional[st.PrivacyLimit],
        salt: Optional[int] = None,
    ) -> Optional[bool]:
        """Check if the constraint attached to a Dataspec meets requirements.

        This function is useful because comparisons are not straightforwards.
        For instance, a Dataspec might have the variant constraint SYNTHETIC
        attached to it. This synthetic dataspec also verifies the DP constraint
        and the PUBLIC constraint.

        Args:
            variant_constraint: VariantConstraint attached to the Dataspec
            kind: constraint kind to verify compliance with
            public_context: actual current public context
            epsilon: current privacy consumed
        """
        ...

    def verified_constraints(
        self, dataspec: st.DataSpec
    ) -> List[st.VariantConstraint]:
        """Return the list of VariantConstraints attached to a DataSpec.

        A VariantConstraint attached to a DataSpec means that the DataSpec
        verifies the constraint.
        """
        ...

    def pep_token(self, dataspec: st.DataSpec) -> Optional[str]:
        """Return a token if the dataspec is PEP, otherwise return None.

        DataSpec.pep_token() returns a PEP token if the dataset is PEP and None
        otherwise. The PEP token is stored in the properties of the
        VariantConstraint. It is a hash initialized with a value when the
        Dataset is protected.

        If a transform does not preserve the PEID then the token is set to None
        If a transform preserves the PEID assignment but changes the rows (e.g.
        sample, shuffle, filter,...) then the token's value is changed If a
        transform does not change the rows (e.g. selecting a column, adding a
        scalar,...) then the token is passed without change

        A Dataspec is PEP if its PEP token is not None. Two PEP Dataspecs are
        aligned (i.e. they have the same number of rows and all their rows have
        the same PEID) if their tokens are equal.
        """
        ...

    def is_public(self, dataspec: st.DataSpec) -> bool:
        """Return True if the dataspec is public.

        Some DataSpecs are intrinsically Public, this is the case if they are
        freely available externally, they can be tagged so and will never be
        considered otherwise.

        This function returns True in the following cases:
        - The dataspec is an ML model
        - The dataspec is transformed but all its inputs are public

        This functions creates a VariantConstraint on the DataSpec to cache the
        PUBLIC constraint.
        """
        ...

    def is_dp(self, dataspec: st.DataSpec) -> bool:
        """Return True if the dataspec is the result of a DP transform.

        This is a simple implementation. This function checks if the
        dataspec's transform has a privacy budget and a random seed as an
        argument.
        """
        ...

    def is_synthetic(self, dataspec: st.DataSpec) -> bool:
        """Return True if the dataspec is synthetic.

        This functions creates a VariantConstraint on the DataSpec to cache
        the SYNTHETIC constraint.
        """

    def private_queries(self, dataspec: st.DataSpec) -> List[st.PrivateQuery]:
        """Return the list of PrivateQueries used in a Dataspec's transform.

        It represents the privacy loss associated with the current computation.

        It can be used by Sarus when a user (Access object) reads a DP dataspec
        to update its accountant. Note that Private Query objects are generated
        with a random uuid so that even if they are submitted multiple times to
        an account, they are only accounted once (ask @cgastaud for more on
        accounting)."""
        ...

    def is_dp_writable(self, dataspec: st.DataSpec) -> bool:
        ...

    def is_pep_writable(self, dataspec: st.DataSpec) -> bool:
        ...

    def is_publishable(self, dataspec: st.DataSpec) -> bool:
        ...

    def is_published(self, dataspec: st.DataSpec) -> bool:
        ...

    def rewritten_pep_token(self, dataspec: st.DataSpec) -> Optional[str]:
        """Returns the PEP token for the DP variant of this dataspec
        during DP rewriting.

        Currently, the implementation assumes a single DP/PEP variant per
        dataspec, resulting in one possible value for
        "rewritten_pep_token."
        Future changes could introduce multiple variants, necessitating
        the implementation of priority rules.
        """
        ...



from __future__ import annotations

from typing import Collection, Optional, cast
import logging

from sarus_data_spec.constants import VARIANT_UUID
from sarus_data_spec.dataspec_validator.base import BaseDataspecValidator
from sarus_data_spec.storage.typing import Storage
import sarus_data_spec.dataspec_rewriter.simple_rules as rewriting_rules
import sarus_data_spec.dataspec_validator.typing as sdvt
import sarus_data_spec.typing as st

logger = logging.getLogger(__name__)


class BaseDataspecRewriter:
    def __init__(self, storage: Storage):
        self._storage = storage
        self._dataspec_validator = BaseDataspecValidator(storage=storage)

    def dataspec_validator(self) -> sdvt.DataspecValidator:
        return self._dataspec_validator

    def storage(self) -> Storage:
        return self._storage

    def variant(
        self,
        dataspec: st.DataSpec,
        kind: st.ConstraintKind,
        public_context: Collection[str],
        privacy_limit: Optional[st.PrivacyLimit],
        salt: Optional[int] = None,
    ) -> Optional[st.DataSpec]:
        return rewriting_rules.rewrite(
            self.dataspec_validator(),
            dataspec,
            kind,
            public_context,
            privacy_limit,
            salt,
        )

    def variants(self, dataspec: st.DataSpec) -> Collection[st.DataSpec]:
        """Return all variants attached to a Dataspec."""
        variants_attributes = {
            variant_kind: dataspec.attributes(name=variant_kind.name)
            for variant_kind in st.ConstraintKind
        }
        variants_dict = {
            variant_kind: [
                self.storage().referrable(att[VARIANT_UUID]) for att in atts
            ]
            for variant_kind, atts in variants_attributes.items()
        }
        # raise warning if some variants are not found in the storage
        for variant_kind, variants in variants_dict.items():
            if any([variant is None for variant in variants]):
                logger.warning(
                    "Found None "
                    f"variant {variant_kind.name} for dataspec {dataspec}"
                )
        variants = list(
            filter(lambda x: x is not None, sum(variants_dict.values(), []))
        )
        return cast(Collection[st.DataSpec], variants)



import typing as t

import pandas as pd
import pyarrow as pa

from sarus_data_spec.arrow.pandas_utils import remove_pandas_index_columns
from sarus_data_spec.json_serialisation import (
    SarusJSONDecoder,
    SarusJSONEncoder,
)
from sarus_data_spec.manager.async_utils import (
    sync_iterator_from_async_iterator,
)
import sarus_data_spec.typing as st

NO_SERIES_NAME = "__sarus_no_name__"
GROUPBY_COLUMN_NAME = "__sarus_groupby_index_"
GROUPBY_KEYS_NAME = "sarus_groupby_keys_names"


async def async_cast_arrow_batches(
    batches_async_iterator: t.AsyncIterator[pa.RecordBatch],
    kind: st.DatasetCastable,
) -> st.DatasetCastable:
    """Convert an async record batches iterator to another Python type."""
    if kind not in t.get_args(st.DatasetCastable):
        raise TypeError(f"Cannot cast a Dataset to type {kind}")

    if kind == t.Iterator[pa.RecordBatch]:
        return sync_iterator_from_async_iterator(batches_async_iterator)
    if kind == pd.DataFrame:
        return await async_arrow_batches_to_dataframe(batches_async_iterator)
    elif kind == pd.Series:
        return await async_arrow_batches_to_series(batches_async_iterator)
    elif kind == pd.core.groupby.DataFrameGroupBy:
        return await async_arrow_batches_to_dataframegroupby(
            batches_async_iterator
        )
    elif kind == pd.core.groupby.SeriesGroupBy:
        return await async_arrow_batches_to_seriesgroupby(
            batches_async_iterator
        )
    else:
        raise NotImplementedError(
            f"Arrow to Python: No converter implemented for type {kind}"
        )


async def async_arrow_batches_to_dataframe(
    batches_async_iterator: t.AsyncIterator[pa.RecordBatch],
) -> pd.DataFrame:
    arrow_batches = [batch async for batch in batches_async_iterator]
    # follow advices in
    # https://arrow.apache.org/docs/python/pandas.html#reducing
    # -memory-use-in-table-to-pandas
    tab = pa.Table.from_batches(arrow_batches)
    df = tab.to_pandas(split_blocks=False, self_destruct=True)
    del tab  # not necessary, but a good practice
    return df


async def async_df_to_series(df: pd.DataFrame) -> pd.Series:
    n_rows, n_cols = df.shape
    if n_cols == 1:
        first_col = df.columns[0]
        series = df[first_col]
        if first_col == NO_SERIES_NAME:
            # This is a default value for a column.
            # The series actually has no name.
            series.name = None
    elif n_rows == 1:
        series = df.iloc[0]
    else:
        raise ValueError(
            "Trying to convert to series Arrow batches "
            "with more than 1 line and 1 column."
        )

    return series


async def async_arrow_batches_to_series(
    batches_async_iterator: t.AsyncIterator[pa.RecordBatch],
) -> pd.Series:
    """Returns the first columns of the DataFrame."""
    df = await async_arrow_batches_to_dataframe(batches_async_iterator)
    series = await async_df_to_series(df)
    return series


async def async_arrow_batches_to_dataframegroupby(
    batches_async_iterator: t.AsyncIterator[pa.RecordBatch],
) -> pd.Series:
    """Returns the group by object from the data."""
    arrow_batches = [batch async for batch in batches_async_iterator]
    tab = pa.Table.from_batches(arrow_batches)

    groupby_keys_bytes = bytes(GROUPBY_KEYS_NAME, 'utf-8')
    if groupby_keys_bytes in tab.schema.metadata:
        keys_encoded = tab.schema.metadata[groupby_keys_bytes]
        keys = SarusJSONDecoder.decode_bytes(keys_encoded)
    else:
        raise ValueError(
            "Trying to convert to series Arrow batches "
            "without groupby keys already defined."
        )

    df = tab.to_pandas(split_blocks=False, self_destruct=True)
    del tab  # not necessary, but a good practice
    groupby_object = df.groupby(keys)
    return groupby_object


async def async_arrow_batches_to_seriesgroupby(
    batches_async_iterator: t.AsyncIterator[pa.RecordBatch],
) -> pd.Series:
    """Returns the group by object from the data."""
    arrow_batches = [batch async for batch in batches_async_iterator]
    tab = pa.Table.from_batches(arrow_batches)
    df = tab.to_pandas(split_blocks=False, self_destruct=True)

    groupby_keys_bytes = bytes(GROUPBY_KEYS_NAME, 'utf-8')
    if groupby_keys_bytes in tab.schema.metadata:
        keys_encoded = tab.schema.metadata[groupby_keys_bytes]
        keys = SarusJSONDecoder.decode_bytes(keys_encoded)
    else:
        raise ValueError(
            "Trying to convert to series Arrow batches "
            "without groupby keys already defined."
        )
    del tab  # not necessary, but a good practice

    series = await async_df_to_series(df)

    groupby_object = series.groupby(keys)
    return groupby_object


def to_pyarrow_table(data: t.Any) -> pa.Table:
    """Convert the result of an external transform to a Pyarrow Table."""
    if not type(data) in t.get_args(st.DatasetCastable):
        raise TypeError(f"Cannot convert {type(data)} to Arrow batches.")

    if isinstance(data, pd.DataFrame):
        df = t.cast(pd.DataFrame, data)
        return pa.Table.from_pandas(df)
    elif isinstance(data, pd.Series):
        sr = t.cast(pd.Series, data)
        if sr.name is None:
            # We need to set a name otherwise pandas adds a default one.
            sr.name = NO_SERIES_NAME
        return pa.Table.from_pandas(pd.DataFrame(sr))
    elif isinstance(data, pd.core.groupby.DataFrameGroupBy):
        df_grouped_by = t.cast(pd.core.groupby.DataFrameGroupBy, data)
        combined_df = df_grouped_by.obj

        keys = df_grouped_by.keys
        encoded_keys = SarusJSONEncoder.encode_bytes(keys)

        table_without_keys = remove_pandas_index_columns(
            pa.Table.from_pandas(combined_df)
        )
        table_with_keys = table_without_keys.replace_schema_metadata(
            {GROUPBY_KEYS_NAME: encoded_keys}
        )

        return table_with_keys
    elif isinstance(data, pd.core.groupby.SeriesGroupBy):
        series_grouped_by = t.cast(pd.core.groupby.SeriesGroupBy, data)
        combined_series = series_grouped_by.obj
        if combined_series.name is None:
            combined_series.name = NO_SERIES_NAME

        keys = df_grouped_by.keys
        encoded_keys = SarusJSONEncoder.encode_bytes(keys)

        table_without_keys = remove_pandas_index_columns(
            pa.Table.from_pandas(pd.DataFrame(combined_series))
        )
        table_with_keys = table_without_keys.replace_schema_metadata(
            {GROUPBY_KEYS_NAME: encoded_keys}
        )

        return table_with_keys
    elif isinstance(data, t.Iterator):
        # We test this case last because DataFrames and Series are also
        # Iterators. We cannot easily test that an object is an Iterator of a
        # specific type. So we put the Iterator[pa.RecordBatch] last as the
        # last possible case.
        batches = t.cast(t.Iterator[pa.RecordBatch], data)
        return pa.Table.from_batches(batches)
    else:
        raise NotImplementedError(
            f"Python to Arrow: No converter implemented for type {type(data)}"
        )



import typing as t
import warnings

import numpy as np
import pyarrow as pa

from sarus_data_spec.arrow.array import convert_record_batch
from sarus_data_spec.bounds import bounds as bounds_builder
from sarus_data_spec.constants import DATA, DATASET_SLUGNAME
from sarus_data_spec.dataset import Dataset
from sarus_data_spec.manager.ops.processor.standard.sample import (
    fast_gather,
    new_sampling_ratio,
    sample_arrow_to_arrow,
)
from sarus_data_spec.manager.ops.processor.standard.standard_op import (  # noqa: E501
    StandardDatasetImplementation,
    StandardDatasetStaticChecker,
)
from sarus_data_spec.marginals import marginals as marginals_builder
from sarus_data_spec.multiplicity import multiplicity as multiplicity_builder
from sarus_data_spec.path import straight_path
from sarus_data_spec.scalar import Scalar
from sarus_data_spec.schema import schema
from sarus_data_spec.size import size as size_builder
import sarus_data_spec.typing as st

try:
    import sqlalchemy as sa
except ModuleNotFoundError:
    warnings.warn('sqlalchemy not installed. No sampling on bigdata')

try:
    from sarus_data_spec.manager.ops.processor.standard.sampling.differentiated_sampling_sizes import (  # noqa: E501
        differentiated_sampling_sizes,
    )
    from sarus_data_spec.manager.ops.processor.standard.sampling.size_utils import (  # noqa: E501
        differentiated_sampled_size,
        sampled_size,
    )
except ModuleNotFoundError as exception:
    # for the public repo
    if (
        exception.name
        == 'sarus_data_spec.manager.ops.processor.standard.sampling'  # noqa: E501
    ):
        pass
    else:
        raise exception

try:
    from sarus_data_spec.manager.ops.sql_utils.bigdata import (
        find_optimal_multiplier_fraction,
        path_to_quoted_string,
        sqlalchemy_query_to_string,
    )
    from sarus_data_spec.manager.ops.sql_utils.queries import nest_queries
    from sarus_data_spec.manager.ops.sql_utils.schema_translations import (
        async_sa_metadata_from_dataset,
    )
except ModuleNotFoundError:
    warnings.warn('sql utils not installed.')


class DifferentiatedSampleStaticChecker(StandardDatasetStaticChecker):
    async def schema(self) -> st.Schema:
        parent_schema = await self.parent_schema()
        return schema(
            self.dataset,
            schema_type=parent_schema.type(),
            protected_paths=parent_schema.protobuf().protected,
            properties=parent_schema.properties(),
            name=self.dataset.properties().get(DATASET_SLUGNAME, None),
        )


class DifferentiatedSample(StandardDatasetImplementation):
    """Computes schema and arrow
    batches for a dataspec transformed by
    a differentiated transform
    """

    async def size(self) -> st.Size:
        parent_size = await self.parent_size()

        previous_schema = await self.parent_schema()
        if len(previous_schema.tables()) == 1:
            size_ratio = new_sampling_ratio(
                self.dataset, parent_size.statistics()
            )
            if size_ratio >= 1:
                return parent_size
            return size_builder(
                self.dataset,
                sampled_size(parent_size.statistics(), size_ratio),
            )
        size_dict = await differentiated_sampling_sizes(self.dataset)
        return size_builder(
            self.dataset,
            differentiated_sampled_size(
                parent_size.statistics(), size_dict, curr_path=[DATA]
            ),
        )

    async def multiplicity(self) -> st.Multiplicity:
        parent_multiplicity = await self.parent_multiplicity()
        return multiplicity_builder(
            self.dataset, parent_multiplicity.statistics()
        )

    async def bounds(self) -> st.Bounds:
        parent_bounds = await self.parent_bounds()
        previous_schema = await self.parent_schema()
        if len(previous_schema.tables()) == 1:
            size_ratio = new_sampling_ratio(
                self.dataset, parent_bounds.statistics()
            )

            if size_ratio >= 1:
                return parent_bounds

            return bounds_builder(
                self.dataset,
                sampled_size(parent_bounds.statistics(), size_ratio),
            )

        size_dict = await differentiated_sampling_sizes(self.dataset)
        return bounds_builder(
            self.dataset,
            differentiated_sampled_size(
                parent_bounds.statistics(), size_dict, curr_path=[DATA]
            ),
        )

    async def marginals(self) -> st.Marginals:
        parent_marginals = await self.parent_marginals()
        previous_schema = await self.parent_schema()
        if len(previous_schema.tables()) == 1:
            size_ratio = new_sampling_ratio(
                self.dataset, parent_marginals.statistics()
            )

            if size_ratio >= 1:
                return parent_marginals

            return marginals_builder(
                self.dataset,
                sampled_size(parent_marginals.statistics(), size_ratio),
            )

        size_dict = await differentiated_sampling_sizes(self.dataset)
        return marginals_builder(
            self.dataset,
            differentiated_sampled_size(
                parent_marginals.statistics(), size_dict, curr_path=[DATA]
            ),
        )

    async def to_arrow(
        self, batch_size: int
    ) -> t.AsyncIterator[pa.RecordBatch]:
        parent = t.cast(Dataset, self.parent())
        if parent.manager().is_big_data(parent):
            queries = await self.sql_implementation()
            assert queries is not None
            # rewrite queries as nested dict rather than dict[tuple,str]
            nested_queries = nest_queries(queries)
            schema = await self.dataset.async_schema()
            return await self.dataset.manager().async_sql(
                parent,
                query=nested_queries,
                batch_size=batch_size,
                result_type=schema.type(),
            )
        return await self._arrow_to_arrow(batch_size)

    async def _arrow_to_arrow(
        self, batch_size: int
    ) -> t.AsyncIterator[pa.RecordBatch]:
        seed = Scalar(
            self.dataset.transform().protobuf().spec.differentiated_sample.seed
        ).value()
        generator = np.random.default_rng(seed)

        previous_ds = t.cast(Dataset, self.parent())
        previous_schema = await self.parent_schema()
        parent_batches = [
            convert_record_batch(batch, previous_schema.type())
            async for batch in await self.parent_to_arrow()
        ]
        struct_arr = pa.concat_arrays(parent_batches)
        previous_size = await previous_ds.manager().async_size(previous_ds)
        assert previous_size

        if len(previous_schema.tables()) == 1:
            return await sample_arrow_to_arrow(self, batch_size)

        size_dict = await differentiated_sampling_sizes(self.dataset)

        indices_to_take = sample_indices_from_array(
            array=struct_arr.flatten()[struct_arr.type.get_field_index(DATA)],
            stat=previous_size.statistics(),
            selected_indices=pa.array(
                np.linspace(0, len(struct_arr) - 1, len(struct_arr), dtype=int)
            ),
            new_size_dict=size_dict,
            curr_path=[DATA],
            random_gen=generator,
        )
        return fast_gather(
            indices=indices_to_take.to_numpy(),
            batches=pa.Table.from_arrays(
                struct_arr.flatten(),
                names=list(previous_schema.type().children().keys()),
            ).to_batches(max_chunksize=1000),
            batch_size=batch_size,
        )

    async def sql_implementation(
        self,
    ) -> t.Optional[t.Dict[t.Tuple[str, ...], str]]:

        schema = await self.parent_schema()
        previous_size = await self.parent_size()
        size_dict = await differentiated_sampling_sizes(self.dataset)
        assert previous_size
        previous_stats = previous_size.statistics()
        sqlalchemy_metadata = await async_sa_metadata_from_dataset(
            self.dataset
        )

        queries = {}
        for table_path in schema.tables():
            curr_path = table_path.to_strings_list()[0]
            full_tablename = (
                curr_path[1:] if curr_path[0] == DATA else curr_path
            )
            sa_table_name = path_to_quoted_string(
                straight_path(full_tablename)
            )
            sa_table = sqlalchemy_metadata.tables[sa_table_name]
            sample_size = size_dict[table_path]
            if (
                self.dataset.transform()
                .protobuf()
                .spec.differentiated_sample.HasField('fraction')
            ):
                fraction = (
                    self.dataset.transform()
                    .protobuf()
                    .spec.differentiated_sample.fraction
                )
                sample_query = sa.select(sa_table).where(
                    sa.func.random() < fraction
                )
            else:
                fraction = (
                    sample_size
                    / previous_stats.nodes_statistics(table_path)[0].size()
                )
                mult_fraction = find_optimal_multiplier_fraction(
                    previous_stats.nodes_statistics(table_path)[0].size(),
                    sample_size,
                )
                optimal_fraction = round(fraction * mult_fraction, 4)
                sample_query = (
                    sa.select(sa_table)
                    .where(
                        sa.func.random()
                        < sa.literal(optimal_fraction, sa.FLOAT)
                    )
                    .order_by(sa.func.random())
                    .limit(sample_size)
                )

            queries[tuple(full_tablename)] = sqlalchemy_query_to_string(
                sample_query
            )

        return queries


def sample_indices_from_array(
    array: pa.Array,
    stat: st.Statistics,
    selected_indices: pa.Array,
    new_size_dict: t.Dict[st.Path, int],
    random_gen: np.random.Generator,
    curr_path: t.List[str],
) -> pa.Array:
    class IndicesSampler(st.StatisticsVisitor):

        indices: pa.Array = selected_indices
        batch_array: pa.Array = array

        def Union(
            self,
            fields: t.Mapping[str, st.Statistics],
            size: int,
            multiplicity: float,
            name: t.Optional[str] = None,
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:

            new_indices = []
            for field, field_stat in fields.items():
                index_element = self.batch_array.type.get_field_index(field)
                filter_idex = pa.array(
                    np.equal(
                        self.batch_array.field('field_selected'),
                        np.array(field),
                    )
                )
                updated_indices = sample_indices_from_array(
                    self.batch_array.filter(filter_idex).flatten()[
                        index_element
                    ],
                    field_stat,
                    self.indices.filter(filter_idex),
                    new_size_dict,
                    random_gen,
                    [*(el for el in curr_path), field],
                )
                new_indices.append(updated_indices)
            self.indices = pa.concat_arrays(new_indices)

        def Struct(
            self,
            fields: t.Mapping[str, st.Statistics],
            size: int,
            multiplicity: float,
            name: t.Optional[str] = None,
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:

            # sample indices to size and sample ratio
            new_size = new_size_dict[straight_path(curr_path.copy())]
            self.indices = pa.array(
                random_gen.choice(
                    self.indices,
                    replace=False,
                    size=min(new_size, size, len(array)),
                )
            )  # need to put array otherwise because of dp
            # we can have size and new size larger and an error will be raised

        def Null(self, size: int, multiplicity: float) -> None:
            raise NotImplementedError

        def Unit(self, size: int, multiplicity: float) -> None:
            raise NotImplementedError

        def Boolean(
            self,
            size: int,
            multiplicity: float,
            probabilities: t.Optional[t.List[float]] = None,
            names: t.Optional[t.List[bool]] = None,
            values: t.Optional[t.List[int]] = None,
        ) -> None:
            raise NotImplementedError

        def Id(self, size: int, multiplicity: float) -> None:
            raise NotImplementedError

        def Integer(
            self,
            size: int,
            multiplicity: float,
            min_value: int,
            max_value: int,
            probabilities: t.Optional[t.List[float]] = None,
            values: t.Optional[t.List[int]] = None,
        ) -> None:
            raise NotImplementedError

        def Enum(
            self,
            size: int,
            multiplicity: float,
            probabilities: t.Optional[t.List[float]] = None,
            names: t.Optional[t.List[str]] = None,
            values: t.Optional[t.List[float]] = None,
            name: str = 'Enum',
        ) -> None:
            raise NotImplementedError

        def Float(
            self,
            size: int,
            multiplicity: float,
            min_value: float,
            max_value: float,
            probabilities: t.Optional[t.List[float]] = None,
            values: t.Optional[t.List[float]] = None,
        ) -> None:
            raise NotImplementedError

        def Text(
            self,
            size: int,
            multiplicity: float,
            min_value: int,
            max_value: int,
            example: str = '',
            probabilities: t.Optional[t.List[float]] = None,
            values: t.Optional[t.List[int]] = None,
        ) -> None:
            raise NotImplementedError

        def Bytes(self, size: int, multiplicity: float) -> None:
            raise NotImplementedError

        def Optional(
            self, statistics: st.Statistics, size: int, multiplicity: float
        ) -> None:
            raise NotImplementedError

        def List(
            self,
            statistics: st.Statistics,
            size: int,
            multiplicity: float,
            min_value: int,
            max_value: int,
            name: str = 'List',
            probabilities: t.Optional[t.List[float]] = None,
            values: t.Optional[t.List[int]] = None,
        ) -> None:
            raise NotImplementedError

        def Array(
            self,
            statistics: st.Statistics,
            size: int,
            multiplicity: float,
            min_values: t.Optional[t.List[float]] = None,
            max_values: t.Optional[t.List[float]] = None,
            name: str = 'Array',
            probabilities: t.Optional[t.List[t.List[float]]] = None,
            values: t.Optional[t.List[t.List[float]]] = None,
        ) -> None:
            raise NotImplementedError

        def Datetime(
            self,
            size: int,
            multiplicity: float,
            min_value: int,
            max_value: int,
            probabilities: t.Optional[t.List[float]] = None,
            values: t.Optional[t.List[int]] = None,
        ) -> None:
            raise NotImplementedError

        def Constrained(
            self, statistics: st.Statistics, size: int, multiplicity: float
        ) -> None:
            raise NotImplementedError

        def Hypothesis(
            self,
            *types: t.Tuple[st.Type, float],
            name: t.Optional[str] = None,
            properties: t.Optional[t.Mapping[str, str]] = None,
        ) -> None:
            raise NotImplementedError

        def Time(
            self,
            size: int,
            multiplicity: float,
            min_value: int,
            max_value: int,
            probabilities: t.Optional[t.List[float]] = None,
            values: t.Optional[t.List[int]] = None,
        ) -> None:
            raise NotImplementedError

        def Date(
            self,
            size: int,
            multiplicity: float,
            min_value: int,
            max_value: int,
            probabilities: t.Optional[t.List[float]] = None,
            values: t.Optional[t.List[int]] = None,
        ) -> None:
            raise NotImplementedError

        def Duration(
            self,
            size: int,
            multiplicity: float,
            min_value: int,
            max_value: int,
            probabilities: t.Optional[t.List[float]] = None,
            values: t.Optional[t.List[int]] = None,
        ) -> None:
            raise NotImplementedError

    visitor = IndicesSampler()
    stat.accept(visitor)
    return visitor.indices



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""
import builtins
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.internal.enum_type_wrapper
import google.protobuf.message
import typing
import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor = ...

class Scalar(google.protobuf.message.Message):
    """A Scalar represents data that does not fulfill the promise of a Dataset.
    A Dataset promises to have a schema and the possibility to iterate on
    pyarrow.RecordBatches
    """
    DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
    class PropertiesEntry(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        key: typing.Text = ...
        value: typing.Text = ...
        def __init__(self,
            *,
            key : typing.Text = ...,
            value : typing.Text = ...,
            ) -> None: ...
        def ClearField(self, field_name: typing_extensions.Literal["key",b"key","value",b"value"]) -> None: ...

    class Spec(google.protobuf.message.Message):
        """Definitions
        How to obtain the dataset
        """
        DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
        TRANSFORMED_FIELD_NUMBER: builtins.int
        MODEL_FIELD_NUMBER: builtins.int
        PRIVACY_PARAMS_FIELD_NUMBER: builtins.int
        RANDOM_SEED_FIELD_NUMBER: builtins.int
        SYNTHETIC_MODEL_FIELD_NUMBER: builtins.int
        @property
        def transformed(self) -> global___Scalar.Transformed: ...
        @property
        def model(self) -> global___Scalar.Model: ...
        @property
        def privacy_params(self) -> global___Scalar.PrivacyParameters: ...
        @property
        def random_seed(self) -> global___Scalar.RandomSeed: ...
        @property
        def synthetic_model(self) -> global___Scalar.SyntheticModel: ...
        def __init__(self,
            *,
            transformed : typing.Optional[global___Scalar.Transformed] = ...,
            model : typing.Optional[global___Scalar.Model] = ...,
            privacy_params : typing.Optional[global___Scalar.PrivacyParameters] = ...,
            random_seed : typing.Optional[global___Scalar.RandomSeed] = ...,
            synthetic_model : typing.Optional[global___Scalar.SyntheticModel] = ...,
            ) -> None: ...
        def HasField(self, field_name: typing_extensions.Literal["model",b"model","privacy_params",b"privacy_params","random_seed",b"random_seed","spec",b"spec","synthetic_model",b"synthetic_model","transformed",b"transformed"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing_extensions.Literal["model",b"model","privacy_params",b"privacy_params","random_seed",b"random_seed","spec",b"spec","synthetic_model",b"synthetic_model","transformed",b"transformed"]) -> None: ...
        def WhichOneof(self, oneof_group: typing_extensions.Literal["spec",b"spec"]) -> typing.Optional[typing_extensions.Literal["transformed","model","privacy_params","random_seed","synthetic_model"]]: ...

    class Transformed(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
        class NamedArgumentsEntry(google.protobuf.message.Message):
            DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
            KEY_FIELD_NUMBER: builtins.int
            VALUE_FIELD_NUMBER: builtins.int
            key: typing.Text = ...
            value: typing.Text = ...
            def __init__(self,
                *,
                key : typing.Text = ...,
                value : typing.Text = ...,
                ) -> None: ...
            def ClearField(self, field_name: typing_extensions.Literal["key",b"key","value",b"value"]) -> None: ...

        TRANSFORM_FIELD_NUMBER: builtins.int
        ARGUMENTS_FIELD_NUMBER: builtins.int
        NAMED_ARGUMENTS_FIELD_NUMBER: builtins.int
        transform: typing.Text = ...
        """Transform id"""

        @property
        def arguments(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[typing.Text]:
            """Dataset or other object ids"""
            pass
        @property
        def named_arguments(self) -> google.protobuf.internal.containers.ScalarMap[typing.Text, typing.Text]: ...
        def __init__(self,
            *,
            transform : typing.Text = ...,
            arguments : typing.Optional[typing.Iterable[typing.Text]] = ...,
            named_arguments : typing.Optional[typing.Mapping[typing.Text, typing.Text]] = ...,
            ) -> None: ...
        def ClearField(self, field_name: typing_extensions.Literal["arguments",b"arguments","named_arguments",b"named_arguments","transform",b"transform"]) -> None: ...

    class Model(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
        class _ModelClass:
            ValueType = typing.NewType('ValueType', builtins.int)
            V: typing_extensions.TypeAlias = ValueType
        class _ModelClassEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[_ModelClass.ValueType], builtins.type):
            DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor = ...
            TF_KERAS: Scalar.Model.ModelClass.ValueType = ...  # 0
            SK_SVC: Scalar.Model.ModelClass.ValueType = ...  # 1
            SK_ONEHOT: Scalar.Model.ModelClass.ValueType = ...  # 2
            SK_PCA: Scalar.Model.ModelClass.ValueType = ...  # 3
            SK_AFFINITY_PROPAGATION: Scalar.Model.ModelClass.ValueType = ...  # 4
            """Cluster"""

            SK_AGGLOMERATIVE_CLUSTERING: Scalar.Model.ModelClass.ValueType = ...  # 5
            SK_BIRCH: Scalar.Model.ModelClass.ValueType = ...  # 6
            SK_DBSCAN: Scalar.Model.ModelClass.ValueType = ...  # 7
            SK_FEATURE_AGGLOMERATION: Scalar.Model.ModelClass.ValueType = ...  # 8
            SK_KMEANS: Scalar.Model.ModelClass.ValueType = ...  # 9
            SK_MINIBATCH_KMEANS: Scalar.Model.ModelClass.ValueType = ...  # 10
            SK_MEAN_SHIFT: Scalar.Model.ModelClass.ValueType = ...  # 11
            SK_OPTICS: Scalar.Model.ModelClass.ValueType = ...  # 12
            SK_SPECTRAL_CLUSTERING: Scalar.Model.ModelClass.ValueType = ...  # 13
            SK_SPECTRAL_BICLUSTERING: Scalar.Model.ModelClass.ValueType = ...  # 14
            SK_SPECTRAL_COCLUSTERING: Scalar.Model.ModelClass.ValueType = ...  # 15
            SK_ADABOOST_CLASSIFIER: Scalar.Model.ModelClass.ValueType = ...  # 60
            """ensemble"""

            SK_ADABOOST_REGRESSOR: Scalar.Model.ModelClass.ValueType = ...  # 61
            SK_BAGGING_CLASSIFIER: Scalar.Model.ModelClass.ValueType = ...  # 62
            SK_BAGGING_REGRESSOR: Scalar.Model.ModelClass.ValueType = ...  # 63
            SK_EXTRA_TREES_REGRESSOR: Scalar.Model.ModelClass.ValueType = ...  # 64
            SK_EXTRA_TREES_CLASSIFIER: Scalar.Model.ModelClass.ValueType = ...  # 65
            SK_GRADIENT_BOOSTING_CLASSIFIER: Scalar.Model.ModelClass.ValueType = ...  # 66
            SK_GRADIENT_BOOSTING_REGRESSOR: Scalar.Model.ModelClass.ValueType = ...  # 67
            SK_ISOLATION_FOREST: Scalar.Model.ModelClass.ValueType = ...  # 68
            SK_RANDOM_FOREST_CLASSIFIER: Scalar.Model.ModelClass.ValueType = ...  # 69
            SK_RANDOM_FOREST_REGRESSOR: Scalar.Model.ModelClass.ValueType = ...  # 70
            SK_RANDOM_TREES_EMBEDDING: Scalar.Model.ModelClass.ValueType = ...  # 71
            SK_STACKING_CLASSIFIER: Scalar.Model.ModelClass.ValueType = ...  # 72
            SK_STACKING_REGRESSOR: Scalar.Model.ModelClass.ValueType = ...  # 73
            SK_VOTING_CLASSIFIER: Scalar.Model.ModelClass.ValueType = ...  # 74
            SK_VOTING_REGRESSOR: Scalar.Model.ModelClass.ValueType = ...  # 75
            SK_HIST_GRADIENT_BOOSTING_REGRESSOR: Scalar.Model.ModelClass.ValueType = ...  # 76
            SK_HIST_GRADIENT_BOOSTING_CLASSIFIER: Scalar.Model.ModelClass.ValueType = ...  # 77
            SK_REPEATED_STRATIFIED_KFOLD: Scalar.Model.ModelClass.ValueType = ...  # 80
            """Model selection"""

            XGB_CLASSIFIER: Scalar.Model.ModelClass.ValueType = ...  # 92
            """XGB"""

            SK_LABEL_ENCODER: Scalar.Model.ModelClass.ValueType = ...  # 98
            SK_KFOLD: Scalar.Model.ModelClass.ValueType = ...  # 99
        class ModelClass(_ModelClass, metaclass=_ModelClassEnumTypeWrapper):
            pass

        TF_KERAS: Scalar.Model.ModelClass.ValueType = ...  # 0
        SK_SVC: Scalar.Model.ModelClass.ValueType = ...  # 1
        SK_ONEHOT: Scalar.Model.ModelClass.ValueType = ...  # 2
        SK_PCA: Scalar.Model.ModelClass.ValueType = ...  # 3
        SK_AFFINITY_PROPAGATION: Scalar.Model.ModelClass.ValueType = ...  # 4
        """Cluster"""

        SK_AGGLOMERATIVE_CLUSTERING: Scalar.Model.ModelClass.ValueType = ...  # 5
        SK_BIRCH: Scalar.Model.ModelClass.ValueType = ...  # 6
        SK_DBSCAN: Scalar.Model.ModelClass.ValueType = ...  # 7
        SK_FEATURE_AGGLOMERATION: Scalar.Model.ModelClass.ValueType = ...  # 8
        SK_KMEANS: Scalar.Model.ModelClass.ValueType = ...  # 9
        SK_MINIBATCH_KMEANS: Scalar.Model.ModelClass.ValueType = ...  # 10
        SK_MEAN_SHIFT: Scalar.Model.ModelClass.ValueType = ...  # 11
        SK_OPTICS: Scalar.Model.ModelClass.ValueType = ...  # 12
        SK_SPECTRAL_CLUSTERING: Scalar.Model.ModelClass.ValueType = ...  # 13
        SK_SPECTRAL_BICLUSTERING: Scalar.Model.ModelClass.ValueType = ...  # 14
        SK_SPECTRAL_COCLUSTERING: Scalar.Model.ModelClass.ValueType = ...  # 15
        SK_ADABOOST_CLASSIFIER: Scalar.Model.ModelClass.ValueType = ...  # 60
        """ensemble"""

        SK_ADABOOST_REGRESSOR: Scalar.Model.ModelClass.ValueType = ...  # 61
        SK_BAGGING_CLASSIFIER: Scalar.Model.ModelClass.ValueType = ...  # 62
        SK_BAGGING_REGRESSOR: Scalar.Model.ModelClass.ValueType = ...  # 63
        SK_EXTRA_TREES_REGRESSOR: Scalar.Model.ModelClass.ValueType = ...  # 64
        SK_EXTRA_TREES_CLASSIFIER: Scalar.Model.ModelClass.ValueType = ...  # 65
        SK_GRADIENT_BOOSTING_CLASSIFIER: Scalar.Model.ModelClass.ValueType = ...  # 66
        SK_GRADIENT_BOOSTING_REGRESSOR: Scalar.Model.ModelClass.ValueType = ...  # 67
        SK_ISOLATION_FOREST: Scalar.Model.ModelClass.ValueType = ...  # 68
        SK_RANDOM_FOREST_CLASSIFIER: Scalar.Model.ModelClass.ValueType = ...  # 69
        SK_RANDOM_FOREST_REGRESSOR: Scalar.Model.ModelClass.ValueType = ...  # 70
        SK_RANDOM_TREES_EMBEDDING: Scalar.Model.ModelClass.ValueType = ...  # 71
        SK_STACKING_CLASSIFIER: Scalar.Model.ModelClass.ValueType = ...  # 72
        SK_STACKING_REGRESSOR: Scalar.Model.ModelClass.ValueType = ...  # 73
        SK_VOTING_CLASSIFIER: Scalar.Model.ModelClass.ValueType = ...  # 74
        SK_VOTING_REGRESSOR: Scalar.Model.ModelClass.ValueType = ...  # 75
        SK_HIST_GRADIENT_BOOSTING_REGRESSOR: Scalar.Model.ModelClass.ValueType = ...  # 76
        SK_HIST_GRADIENT_BOOSTING_CLASSIFIER: Scalar.Model.ModelClass.ValueType = ...  # 77
        SK_REPEATED_STRATIFIED_KFOLD: Scalar.Model.ModelClass.ValueType = ...  # 80
        """Model selection"""

        XGB_CLASSIFIER: Scalar.Model.ModelClass.ValueType = ...  # 92
        """XGB"""

        SK_LABEL_ENCODER: Scalar.Model.ModelClass.ValueType = ...  # 98
        SK_KFOLD: Scalar.Model.ModelClass.ValueType = ...  # 99

        ARGUMENTS_FIELD_NUMBER: builtins.int
        NAMED_ARGUMENTS_FIELD_NUMBER: builtins.int
        MODEL_CLASS_FIELD_NUMBER: builtins.int
        arguments: builtins.bytes = ...
        named_arguments: builtins.bytes = ...
        model_class: global___Scalar.Model.ModelClass.ValueType = ...
        def __init__(self,
            *,
            arguments : builtins.bytes = ...,
            named_arguments : builtins.bytes = ...,
            model_class : global___Scalar.Model.ModelClass.ValueType = ...,
            ) -> None: ...
        def ClearField(self, field_name: typing_extensions.Literal["arguments",b"arguments","model_class",b"model_class","named_arguments",b"named_arguments"]) -> None: ...

    class PrivacyParameters(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
        class Point(google.protobuf.message.Message):
            DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
            EPSILON_FIELD_NUMBER: builtins.int
            DELTA_FIELD_NUMBER: builtins.int
            epsilon: builtins.float = ...
            delta: builtins.float = ...
            def __init__(self,
                *,
                epsilon : builtins.float = ...,
                delta : builtins.float = ...,
                ) -> None: ...
            def ClearField(self, field_name: typing_extensions.Literal["delta",b"delta","epsilon",b"epsilon"]) -> None: ...

        POINTS_FIELD_NUMBER: builtins.int
        @property
        def points(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___Scalar.PrivacyParameters.Point]: ...
        def __init__(self,
            *,
            points : typing.Optional[typing.Iterable[global___Scalar.PrivacyParameters.Point]] = ...,
            ) -> None: ...
        def ClearField(self, field_name: typing_extensions.Literal["points",b"points"]) -> None: ...

    class RandomSeed(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
        VALUE_FIELD_NUMBER: builtins.int
        value: builtins.int = ...
        def __init__(self,
            *,
            value : builtins.int = ...,
            ) -> None: ...
        def ClearField(self, field_name: typing_extensions.Literal["value",b"value"]) -> None: ...

    class SyntheticModel(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
        def __init__(self,
            ) -> None: ...

    UUID_FIELD_NUMBER: builtins.int
    NAME_FIELD_NUMBER: builtins.int
    DOC_FIELD_NUMBER: builtins.int
    SPEC_FIELD_NUMBER: builtins.int
    PROPERTIES_FIELD_NUMBER: builtins.int
    uuid: typing.Text = ...
    """A Scalar does not ensure this possibility. As a consequence, oprations
    from standard libraries are allowed (pandas.mean, numpy.std,...) but
    operations implemented for Datasets by Sarus like computing marginals or
    fitting a Keras model cannot be performed on a Scalar.

    Scalars are generated by transforms that explicitly require a specific
    format (e.g. as_pandas, as_numpy,...) or as byproducts of transforms
    (model weights, training history,...).

    e.g. RFC 4122 id used to refer to the dataset (content linked?)
    """

    name: typing.Text = ...
    doc: typing.Text = ...
    @property
    def spec(self) -> global___Scalar.Spec: ...
    @property
    def properties(self) -> google.protobuf.internal.containers.ScalarMap[typing.Text, typing.Text]:
        """Other properties"""
        pass
    def __init__(self,
        *,
        uuid : typing.Text = ...,
        name : typing.Text = ...,
        doc : typing.Text = ...,
        spec : typing.Optional[global___Scalar.Spec] = ...,
        properties : typing.Optional[typing.Mapping[typing.Text, typing.Text]] = ...,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["spec",b"spec"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["doc",b"doc","name",b"name","properties",b"properties","spec",b"spec","uuid",b"uuid"]) -> None: ...
global___Scalar = Scalar



from typing import Any, Optional, Union

import numpy.typing as npt

from sarus_data_spec.dataspec_validator.signature import (
    SarusParameter,
    SarusSignature,
    SarusSignatureValue,
)

from ..external_op import ExternalOpImplementation

try:
    from scipy.sparse import spmatrix
    from sklearn.base import BaseEstimator
except ModuleNotFoundError:
    BaseEstimator = Any
    spmatrix = Any


class sk_fit(ExternalOpImplementation):
    _transform_id = "sklearn.SK_FIT"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=BaseEstimator,
        ),
        SarusParameter(
            name="X",
            annotation=Union[npt.ArrayLike, spmatrix],
        ),
        SarusParameter(
            name="y",
            annotation=Optional[Union[npt.ArrayLike, spmatrix]],
            default=None,
        ),
        SarusParameter(
            name="sample_weight",
            annotation=Optional[npt.ArrayLike],
            default=None,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, kwargs = signature.collect_kwargs_method()
        if kwargs["sample_weight"] is None:
            del kwargs["sample_weight"]
        fitted_model = this.fit(**kwargs)
        return fitted_model


class sk_fit_y(ExternalOpImplementation):
    _transform_id = "sklearn.SK_FIT_Y"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=BaseEstimator,
        ),
        SarusParameter(
            name="y",
            annotation=Union[npt.ArrayLike, spmatrix],
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, kwargs = signature.collect_kwargs_method()
        fitted_model = this.fit(**kwargs)
        return fitted_model


class sk_predict(ExternalOpImplementation):
    _transform_id = "sklearn.SK_PREDICT"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=BaseEstimator,
        ),
        SarusParameter(
            name="X",
            annotation=Union[npt.ArrayLike, spmatrix],
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, kwargs = signature.collect_kwargs_method()
        return this.predict(**kwargs)


class sk_predict_callable(ExternalOpImplementation):
    _transform_id = "sklearn.SK_PREDICT_CALLABLE"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=BaseEstimator,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        (this,) = signature.collect_args()
        return this.predict


class sk_predict_log_proba(ExternalOpImplementation):
    _transform_id = "sklearn.SK_PREDICT_LOG_PROBA"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=BaseEstimator,
        ),
        SarusParameter(
            name="X",
            annotation=Union[npt.ArrayLike, spmatrix],
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, kwargs = signature.collect_kwargs_method()
        return this.predict_log_proba(**kwargs)


class sk_predict_log_proba_callable(ExternalOpImplementation):
    _transform_id = "sklearn.SK_PREDICT_LOG_PROBA_CALLABLE"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=BaseEstimator,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        (this,) = signature.collect_args()
        return this.predict_log_proba


class sk_predict_proba(ExternalOpImplementation):
    _transform_id = "sklearn.SK_PREDICT_PROBA"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=BaseEstimator,
        ),
        SarusParameter(
            name="X",
            annotation=Union[npt.ArrayLike, spmatrix],
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, kwargs = signature.collect_kwargs_method()
        return this.predict_proba(**kwargs)


class sk_predict_proba_callable(ExternalOpImplementation):
    _transform_id = "sklearn.SK_PREDICT_PROBA_CALLABLE"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=BaseEstimator,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        (this,) = signature.collect_args()
        return this.predict_proba


class sk_transform(ExternalOpImplementation):
    _transform_id = "sklearn.SK_TRANSFORM"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=BaseEstimator,
        ),
        SarusParameter(
            name="X",
            annotation=Union[npt.ArrayLike, spmatrix],
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        (this, X) = signature.collect_args()
        return this.transform(X)


class sk_inverse_transform(ExternalOpImplementation):
    _transform_id = "sklearn.SK_INVERSE_TRANSFORM"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=BaseEstimator,
        ),
        SarusParameter(
            name="X",
            annotation=Union[npt.ArrayLike, spmatrix],
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, X = signature.collect_args()
        return this.inverse_transform(X)


class sk_pipeline_fit(ExternalOpImplementation):
    _transform_id = "sklearn.SK_PIPELINE_FIT"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=BaseEstimator,
        ),
        SarusParameter(
            name="X",
            annotation=Union[npt.ArrayLike, spmatrix],
        ),
        SarusParameter(
            name="y",
            annotation=Optional[Union[npt.ArrayLike, spmatrix]],
            default=None,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, kwargs = signature.collect_kwargs_method()
        fitted_model = this.fit(**kwargs)
        return fitted_model


class sk_pipeline_fit_transform(ExternalOpImplementation):
    _transform_id = "sklearn.SK_PIPELINE_FIT_TRANSFORM"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=BaseEstimator,
        ),
        SarusParameter(
            name="X",
            annotation=Union[npt.ArrayLike, spmatrix],
        ),
        SarusParameter(
            name="y",
            annotation=Optional[npt.ArrayLike],
            default=None,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, kwargs = signature.collect_kwargs_method()
        return this.fit_transform(**kwargs)


class sk_pipeline_transform(ExternalOpImplementation):
    _transform_id = "sklearn.SK_PIPELINE_TRANSFORM"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=BaseEstimator,
        ),
        SarusParameter(
            name="X",
            annotation=Union[npt.ArrayLike, spmatrix],
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        (this, X) = signature.collect_args()
        return this.transform(X)


class sk_score(ExternalOpImplementation):
    _transform_id = "sklearn.SK_SCORE"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=BaseEstimator,
        ),
        SarusParameter(
            name="X",
            annotation=Union[npt.ArrayLike, spmatrix],
        ),
        SarusParameter(
            name="y",
            annotation=Optional[npt.ArrayLike],
            default=None,
        ),
        SarusParameter(
            name="sample_weight",
            annotation=Optional[npt.ArrayLike],
            default=None,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, kwargs = signature.collect_kwargs_method()
        return this.score(**kwargs)


class sk_fit_transform(ExternalOpImplementation):
    _transform_id = "sklearn.SK_LABEL_ENCODER_FIT_TRANSFORM"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=BaseEstimator,
        ),
        SarusParameter(
            name="X",
            annotation=Union[npt.ArrayLike, spmatrix],
        ),
        SarusParameter(
            name="y",
            annotation=Optional[npt.ArrayLike],
            default=None,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, kwargs = signature.collect_kwargs_method()
        return this.fit_transform(**kwargs)


class sk_split(ExternalOpImplementation):
    _transform_id = "sklearn.SK_SPLIT"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=BaseEstimator,
        ),
        SarusParameter(
            name="X",
            annotation=Union[npt.ArrayLike, spmatrix],
        ),
        SarusParameter(
            name="y",
            annotation=Optional[npt.ArrayLike],
            default=None,
        ),
        SarusParameter(
            name="groups",
            annotation=Optional[npt.ArrayLike],
            default=None,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, kwargs = signature.collect_kwargs_method()
        return this.split(**kwargs)


class sk_get_n_splits(ExternalOpImplementation):
    _transform_id = "sklearn.SK_GET_N_SPLITS"
    _signature = SarusSignature(
        SarusParameter(
            name="this",
            annotation=BaseEstimator,
        ),
    )

    def call(self, signature: SarusSignatureValue) -> Any:
        this, kwargs = signature.collect_kwargs_method()
        return this.get_n_splits(**kwargs)
